import torch
import torch.nn as nn
import torch.nn.functional as F


class CellPathwayPoolingAggregator(nn.Module):
    def __init__(self, cellpathway_indices, pooling_type="mean"):
        """
        Initializes the CellPathwayAggregatorPooling module.
        :param cellpathway_indices: A list of lists, each sublist contains indices of gene sets for a specific cell pathway.
        :param pooling_type: A string indicating the type of pooling ('mean' or 'max').
        """
        super(CellPathwayPoolingAggregator, self).__init__()
        self.cellpathway_indices = cellpathway_indices
        self.pooling_type = pooling_type

    def forward(self, gene_set_features):
        """
        Forward pass of the module.
        :param gene_set_features: A tensor of shape (batch_size, num_gene_sets, feature_dim).
        :return: A tensor of shape (batch_size, num_cellpathways).
        """
        device = gene_set_features.device
        batch_size = gene_set_features.size(0)
        num_cellpathways = len(self.cellpathway_indices)
        aggregated_features = torch.zeros(batch_size, num_cellpathways, device=device)

        for i, indices in enumerate(self.cellpathway_indices):
            if self.pooling_type == "mean":
                aggregated_features[:, i] = torch.mean(
                    gene_set_features[:, indices], dim=1
                )
            elif self.pooling_type == "max":
                aggregated_features[:, i] = torch.max(
                    gene_set_features[:, indices], dim=1
                )[0]

        return aggregated_features


class CellPathwayAttentionAggregator(nn.Module):
    def __init__(self, cellpathway_indices, softmax_mean=True):
        """
        Initializes the CellPathwayAttentionAggregator module.

        :param cellpathway_indices: A list of lists, each list contains indices of cell types/pathways in a cellpathway set.
        """
        super(CellPathwayAttentionAggregator, self).__init__()
        self.cellpathway_indices = cellpathway_indices
        self.softmax_mean = softmax_mean

        # Attention weights for each cell type/pathway in each cellpathway set
        self.attention_weights = nn.ParameterDict(
            {
                f"cellpathway_{i}": nn.Parameter(torch.randn(len(cellpathway), 1))
                for i, cellpathway in enumerate(cellpathway_indices)
            }
        )

    def forward(self, geneset_features):
        """
        Forward pass of the module. Aggregates cell type/pathway level features to cellpathway set level features using attention mechanism.
        :param geneset_features:
        :return: A tensor of shape (batch_size, num_cellpathway_sets), representing aggregated cellpathway set level features.
        """
        batch_size = geneset_features.size(0)
        num_cellpathway_sets = len(self.cellpathway_indices)
        aggregated_features = torch.zeros(
            batch_size, num_cellpathway_sets, device=geneset_features.device
        )

        for i, cellpathway in enumerate(self.cellpathway_indices):
            set_features = geneset_features[:, cellpathway]
            if self.softmax_mean:
                attention_scores = F.softmax(
                    self.attention_weights[f"cellpathway_{i}"], dim=0
                )
                weighted_features = set_features * attention_scores.T
                aggregated_features[:, i] = torch.sum(weighted_features, dim=1)
            else:
                attention_scores = self.attention_weights[f"cellpathway_{i}"]
                weighted_features = set_features * attention_scores.T
                aggregated_features[:, i] = torch.mean(weighted_features, dim=1)
        return aggregated_features




class CellPathwayAggregator(nn.Module):
    def __init__(self, cellpathway_indices, mode="pooling", pooling_type="mean"):
        """
        Initializes the CellPathwayAggregator module.
        :param cellpathway_indices: A list of lists, each sublist contains indices of gene sets for a specific cell pathway.
        :param mode: The aggregation mode ('pooling' or 'attention').
        :param pooling_type: The type of pooling ('mean' or 'max'), used if mode is 'pooling'.
        """
        super(CellPathwayAggregator, self).__init__()
        self.mode = mode

        if mode == "pooling":
            self.aggregator = CellPathwayPoolingAggregator(
                cellpathway_indices, pooling_type
            )
        elif mode == "attention":
            self.aggregator = CellPathwayAttentionAggregator(cellpathway_indices)
        else:
            raise ValueError("Invalid mode. Use 'pooling' or 'attention'.")

    def forward(self, geneset_features):
        """
        Forward pass of the module.
        :param gene_set_features: A tensor of shape (batch_size, num_gene_sets).
        :return: A tensor of shape (batch_size, num_cellpathways).
        """
        # print(geneset_features.device)

        cellpathway_features = self.aggregator(geneset_features)

        # print(cellpathway_features.device)
        return cellpathway_features



from .cosformer import CosformerLayer
from .performer import PerformerLayer
from .transformer import VanillaTransformerLayer
from .flowformer import FlowformerLayer
from .flashformer import FlashTransformerLayer



import plotly.graph_objects as go
import plotly.io as pio


import torch.nn.functional as F

import os
from tqdm import tqdm
from itertools import chain
import pandas as pd
import numpy as np
import random, torch
import matplotlib.pyplot as plt
import seaborn as sns

sns.set(style="white", font_scale=1.3)
import warnings

warnings.filterwarnings("ignore")


from ..tokenizer import CONCEPT, CONCEPT_palette


CELLTYPE_palette = (
    pd.DataFrame([CONCEPT_palette])
    .T.reset_index()
    .sort_index(ascending=False)
    .set_index("index")[0]
    .to_dict()
)
GENESET_palette = CONCEPT.BroadCelltypePathway.map(CELLTYPE_palette).to_dict()
features = pd.Series(CELLTYPE_palette).index.tolist()


def plot_sankey_diagram(
    model,
    concept2plot=["NKcell"],
    reverse=False,
    scale_imp=False,
    title_text="",
    concept_imp_dict=dict(zip(features, [1 for i in features])),
    topK_vis=10,
    font_size=15,
    width=1200,
    height=600,
    margin=dict(l=10, r=1, t=50, b=10),
    **layout_args
):

    try:
        gene_name = model.feature_name
    except:
        gene_name = model.pretrainer.feature_name

    concept_lineage_map = (
        CONCEPT.drop_duplicates(["BroadCelltypePathway", "Lineage"])
        .set_index("BroadCelltypePathway")
        .Lineage
    )
    LINEA_palette = {
        "Lymphoid_lineage_Bcell": "#80ff00",
        "Lymphoid_lineage_T/NKcell": "#0000ff",
        "Myeloid_lineage": "#ff00ff",
        "Mesenchymal_lineage": "#ffff00",
        "Functional_group": "#ff8000",
    }

    init_gene = "genes"
    init_color = "#0362fc"
    gene_color = "#9e9d93"  #'#eeeee4'
    GENE_palette = dict(zip(gene_name, [gene_color for i in gene_name]))

    # all-in-one-colors
    color_dict = {init_gene: init_color}
    color_dict.update(GENE_palette)
    color_dict.update(GENESET_palette)
    color_dict.update(CELLTYPE_palette)
    color_dict.update(LINEA_palette)

    geneset_weights = (
        model.model.latentprojector.cellpathwayprojector.cellpathway_aggregator.aggregator.attention_weights
    )
    gene_weights = (
        model.model.latentprojector.genesetprojector.geneset_aggregator.aggregator.attention_weights
    )

    concept_name = model.model.latentprojector.cellpathwayprojector.cellpathway_names

    concept_idx = model.model.latentprojector.cellpathwayprojector.cellpathway_indices

    geneset_name = model.model.latentprojector.genesetprojector.genesets_names
    geneset_idx = model.model.latentprojector.genesetprojector.genesets_indices

    all_res = []
    for i, (concept, idx) in enumerate(zip(concept_name, concept_idx)):

        # concept, idx,i
        concept_fi = concept_imp_dict[concept]
        geneset = geneset_name[idx]

        if scale_imp:
            n_genesets = len(geneset)
            n_genes = sum([len(geneset_idx[x]) for x in idx])
            concept_fi = concept_fi * n_genesets  # *n_genes
            # print(concept_fi, n_genesets, n_genes )

        ## geneset->celltype
        geneset_w = geneset_weights["cellpathway_%s" % i].cpu().detach()
        geneset_w = F.softmax(geneset_w, dim=0) * concept_fi
        geneset_w = geneset_w.numpy().reshape(
            -1,
        )
        df2 = pd.DataFrame([geneset, [concept for i in idx], geneset_w]).T
        df2.columns = ["source", "target", "weights"]
        df2["group"] = "geneset->celltype"

        ## gene->geneset
        res = []
        for j, fi in zip(idx, geneset_w):

            target = geneset_name[j]
            sources = gene_name[geneset_idx[j]]

            # print(len(sources))
            if scale_imp:
                fi = fi * len(sources)

            gene_w = gene_weights["geneset_%s" % j].cpu().detach()
            gene_w = F.softmax(gene_w, dim=0)
            gene_w = (
                gene_w.numpy().reshape(
                    -1,
                )
                * fi
            )

            df3 = pd.DataFrame([sources, [target for i in sources], gene_w]).T
            df3.columns = ["source", "target", "weights"]
            df3["group"] = "gene->geneset"
            res.append(df3)
        df3 = pd.concat(res)

        ## concat all
        df = pd.concat(
            [
                df2,
                df3,
            ]
        )  # df1, df4]
        df["concept"] = concept
        df["source_color"] = df["source"].map(color_dict)
        df["target_color"] = df["target"].map(color_dict)
        all_res.append(df)

    dfa = pd.concat(all_res)

    ## start to plot the sankey diagram
    df = dfa[
        dfa.concept.isin(concept2plot)
    ]  # 'Treg','Cytotoxic_Tcell', 'Exhausted_Tcell',

    all_labels = list(set(df.source.unique()) | set(df.target.unique()))
    labels = (
        df.groupby("source").weights.sum().sort_values(ascending=False).index.to_list()
    )
    labels.extend(list(set(all_labels) - set(labels)))

    tail_nodes = (
        df.groupby(["group", "source"])
        .weights.sum()
        .sort_values(ascending=False)
        .loc["gene->geneset"]
        .iloc[topK_vis:]
        .index
    )
    filtered_labels = ["" if label in tail_nodes else label for label in labels]

    colors = pd.Series(labels).map(color_dict).to_list()
    label_num_dict = dict(zip(labels, range(len(labels))))

    # Define the source and target indices
    source = df.source.map(label_num_dict).to_list()
    target = df.target.map(label_num_dict).to_list()
    values = df.weights.to_list()

    df1 = df.groupby(["source", "target"])["weights"].mean().reset_index()

    source = df1.source.map(label_num_dict).to_list()
    target = df1.target.map(label_num_dict).to_list()
    values = df1.weights.to_list()

    if title_text is None:
        title_text = "_".join(concept2plot)

    # Create the Sankey diagram
    fig = go.Figure(
        go.Sankey(
            node=dict(
                pad=15,
                thickness=20,
                line=dict(color="black", width=0.01),
                label=filtered_labels,
                color=colors,
            ),
            link=dict(
                source=target if reverse else source,
                target=source if reverse else target,
                value=values,
            ),
        )
    )

    fig.update_layout(
        title_text=title_text,
        title_font_size=30,
        font_size=font_size,
        width=width,
        height=height,
        margin=margin,
        **layout_args
    )
    return df, fig


def get_projector_weights(model, concept2plot=features):

    try:
        gene_name = model.feature_name
    except:
        gene_name = model.pretrainer.feature_name

    concept_lineage_map = (
        CONCEPT.drop_duplicates(["BroadCelltypePathway", "Lineage"])
        .set_index("BroadCelltypePathway")
        .Lineage
    )
    LINEA_palette = {
        "Lymphoid_lineage_Bcell": "#80ff00",
        "Lymphoid_lineage_T/NKcell": "#0000ff",
        "Myeloid_lineage": "#ff00ff",
        "Mesenchymal_lineage": "#ffff00",
        "Functional_group": "#ff8000",
    }

    init_gene = "genes"
    init_color = "#0362fc"
    gene_color = "#9e9d93"  #'#eeeee4'
    GENE_palette = dict(zip(gene_name, [gene_color for i in gene_name]))

    # all-in-one-colors
    color_dict = {init_gene: init_color}
    color_dict.update(GENE_palette)
    color_dict.update(GENESET_palette)
    color_dict.update(CELLTYPE_palette)
    color_dict.update(LINEA_palette)

    geneset_weights = (
        model.model.latentprojector.cellpathwayprojector.cellpathway_aggregator.aggregator.attention_weights
    )
    gene_weights = (
        model.model.latentprojector.genesetprojector.geneset_aggregator.aggregator.attention_weights
    )

    concept_name = model.model.latentprojector.cellpathwayprojector.cellpathway_names

    concept_idx = model.model.latentprojector.cellpathwayprojector.cellpathway_indices

    geneset_name = model.model.latentprojector.genesetprojector.genesets_names
    geneset_idx = model.model.latentprojector.genesetprojector.genesets_indices

    all_res = []
    for i, (concept, idx) in enumerate(zip(concept_name, concept_idx)):

        # concept, idx,i

        geneset = geneset_name[idx]

        ## geneset->celltype
        geneset_w = geneset_weights["cellpathway_%s" % i].cpu().detach()
        geneset_w = F.softmax(geneset_w, dim=0)
        geneset_w = geneset_w.numpy().reshape(
            -1,
        )
        df2 = pd.DataFrame([geneset, [concept for i in idx], geneset_w]).T
        df2.columns = ["source", "target", "weights"]
        df2["group"] = "geneset->celltype"

        ## gene->geneset
        res = []
        for j, fi in zip(idx, geneset_w):

            target = geneset_name[j]
            sources = gene_name[geneset_idx[j]]

            gene_w = gene_weights["geneset_%s" % j].cpu().detach()
            gene_w = F.softmax(gene_w, dim=0)
            gene_w = gene_w.numpy().reshape(
                -1,
            )

            df3 = pd.DataFrame([sources, [target for i in sources], gene_w]).T
            df3.columns = ["source", "target", "weights"]
            df3["group"] = "gene->geneset"
            res.append(df3)
        df3 = pd.concat(res)

        ## concat all
        df = pd.concat(
            [
                df2,
                df3,
            ]
        )  # df1, df4]
        df["concept"] = concept
        df["source_color"] = df["source"].map(color_dict)
        df["target_color"] = df["target"].map(color_dict)
        all_res.append(df)

    dfa = pd.concat(all_res)

    ## start to plot the sankey diagram
    df = dfa[
        dfa.concept.isin(concept2plot)
    ]  # 'Treg','Cytotoxic_Tcell', 'Exhausted_Tcell',
    return df



import torch
import torch.nn.functional as F
import numpy as np

from torch import Tensor
from typing import Optional
from torch import nn
from .norm import create_norm
from .transformer import AbstractTrasnformerLayer


class CosformerAttention(nn.Module):
    """
    cosformer attention in "cosFormer: Rethinking Softmax In Attention"
    https://arxiv.org/abs/2202.08791
    """

    def __init__(
        self,
        embed_dim,
        num_heads,
        kdim=None,
        vdim=None,
        dropout_rate=0.0,
        causal=False,
        has_outproj=True,
        act_fun="gelu",
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if kdim is not None else embed_dim
        self.num_heads = num_heads
        self.has_outproj = has_outproj
        self.act_fun = self.get_act_fun(act_fun)
        # q, k, v projection
        self.k_proj = nn.Linear(self.kdim, embed_dim)
        self.v_proj = nn.Linear(self.vdim, embed_dim)
        self.q_proj = nn.Linear(embed_dim, embed_dim)
        # outprojection
        self.out_proj = nn.Linear(embed_dim, embed_dim)
        # dropout rate
        self.dropout_rate = dropout_rate
        self.attn_dropout = nn.Dropout(dropout_rate)
        # causal
        self.causal = causal

        assert (
            self.embed_dim % self.num_heads == 0
        ), "embed_dim must be divisible by num_heads"

    def get_index(self, seq_len):
        index = np.pi / 2 * torch.arange(1, seq_len + 1).reshape(1, -1, 1)

        return nn.Parameter(index, requires_grad=False)

    def get_act_fun(self, act_fun):
        if act_fun == "relu":
            return F.relu
        elif act_fun == "elu":
            return 1 + F.elu
        elif act_fun == "gelu":
            return F.gelu
        else:
            raise ValueError(f"Unrecognized activation function: {act_fun}.")

    def forward(
        self,
        query: Tensor,
        key: Optional[Tensor] = None,
        value: Optional[Tensor] = None,
        attn_mask: Optional[Tensor] = None,
        eps: Optional[float] = 1e-6,
        output_attentions: bool = False,
    ):
        """Input shape: Sequence x Batch x Embedding
        Args:
            query (Tensor): `(L, N, E)` where L is the target sequence length, N is the batch size,
            E is the embedding dimension.
            key (Tensor): `(S, N, E)` where S is the source sequence length, N is the batch size,
            E is the embedding dimension.
            value (Tensor): `(S, N, E)` where S is the source sequence length, N is the batch size,
            E is the embedding dimension.
            attn_mask (Optional[Tensor], optional): typically used to implement causal attention,
            where the mask prevents the attention from looking forward in time (default: None).
        """
        if key == None:
            key = query
        if value == None:
            value = query

        num_heads = self.num_heads
        tgt_len, bsz, embed_dim = query.size()
        src_len = key.size(0)
        head_dim = embed_dim // num_heads

        # get q, k, v
        # (L, N, E)
        q = self.q_proj(query)
        # (S, N, E)
        k = self.k_proj(key)
        # (S, N, E)
        v = self.v_proj(value)

        # activation
        q = self.act_fun(q)
        k = self.act_fun(k)

        # multihead reshape
        # (N * h, L, d)
        q = q.contiguous().view(-1, bsz * num_heads, head_dim).transpose(0, 1)
        # (N * h, S, d)
        k = k.contiguous().view(-1, bsz * num_heads, head_dim).transpose(0, 1)
        # (N * h, S, d)
        v = v.contiguous().view(-1, bsz * num_heads, head_dim).transpose(0, 1)

        # cos transform
        m = max(src_len, tgt_len)
        # get index and send to cuda
        weight_index = self.get_index(m).to(q)
        # (N * h, L, 2 * d)
        q_ = torch.cat(
            [
                q * torch.sin(weight_index[:, :tgt_len, :] / m),
                q * torch.cos(weight_index[:, :tgt_len, :] / m),
            ],
            dim=-1,
        )
        # (N * h, S, 2 * d)
        k_ = torch.cat(
            [
                k * torch.sin(weight_index[:, :src_len, :] / m),
                k * torch.cos(weight_index[:, :src_len, :] / m),
            ],
            dim=-1,
        )

        if self.causal:
            kv_ = torch.einsum("nld,nlm->nldm", k_, v)
            kv_ = self.attn_dropout(kv_)
            kv_cum = torch.cumsum(kv_, dim=1)
            qkv = torch.einsum("nld,nldm->nlm", q_, kv_cum)
            k_cum = torch.cumsum(k_, dim=1)
            denom = torch.clamp_min(torch.einsum("nlm,nlm->nl", q_, k_cum), eps)
            attn_output = qkv / denom.unsqueeze(-1)
            attn_output = (
                attn_output.transpose(0, 1).contiguous().view(tgt_len, bsz, -1)
            )
            attn_weights = None
        else:
            kv_ = torch.einsum("nld,nlm->ndm", k_, v)
            kv_ = self.attn_dropout(kv_)
            z_ = 1 / torch.clamp_min(
                torch.einsum("nld,nd->nl", q_, torch.sum(k_, axis=1)), eps
            )
            attn_output = torch.einsum("nld,ndm,nl->nlm", q_, kv_, z_)
            attn_output = (
                attn_output.transpose(0, 1).contiguous().view(tgt_len, bsz, -1)
            )

            # 这一步是新增的，用于计算注意力权重
            if output_attentions:
                attn_weights = torch.einsum("nld,ndm->nlm", q_, kv_)
                attn_weights = torch.softmax(attn_weights, dim=-1)

        if self.has_outproj:
            attn_output = self.out_proj(attn_output)

        if output_attentions:
            return attn_output, attn_weights
        else:
            return attn_output


class CosformerLayer(nn.Module, AbstractTrasnformerLayer):
    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        norm="layernorm",
        norm_first: bool = True,
        causal=False,
    ):
        super().__init__()
        self.self_attn = CosformerAttention(embed_dim=embed_dim, num_heads=num_heads)
        self._ff_block = nn.Sequential(
            nn.Linear(embed_dim, embed_dim * 2),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(embed_dim * 2, embed_dim),
            nn.Dropout(dropout),
        )
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = create_norm(norm, embed_dim)
        self.norm2 = create_norm(norm, embed_dim)
        self.norm_first = norm_first
        self.support_output_attentions = False

    def _sa_block(self, x: Tensor, attn_mask: Optional[Tensor]):
        # x = x.unsqueeze(1)
        x = self.self_attn(x, attn_mask=attn_mask)
        return self.dropout1(x)  # [:, 0, :]

    def forward(self, x_blc, attn_mask=None, output_attentions=False):
        """
        x_blc: BLC

        """
        x = x_blc.transpose(0, 1)  # LBC

        assert (
            output_attentions == False
        ), "output_attentions not implemented for Cosformer"
        if self.norm_first:
            x = x + self._sa_block(self.norm1(x), attn_mask)
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(x + self._sa_block(x, attn_mask))
            x = self.norm2(x + self._ff_block(x))

        x_out = x.transpose(0, 1)

        return x_out, None



import torch.nn as nn
import torch
import random


class RMSNorm(nn.Module):
    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x):
        output = self._norm(x.float()).type_as(x)
        return output * self.weight


class DSBNNorm(nn.Module):
    def __init__(
        self,
        dim: int,
        domain_num: int,
        domain_label: str = "dataset",
        eps: float = 1e-6,
        flip_rate=0.3,
    ):
        super().__init__()
        self.eps = eps
        self.domain_label = domain_label
        self.bns = nn.ModuleList([nn.BatchNorm1d(dim) for _ in range(domain_num + 1)])
        self.flip_rate = flip_rate

    def forward(self, xdict):
        h = xdict["h"]
        if self.training and random.random() < self.flip_rate:
            for i in xdict[self.domain_label].unique():
                h[xdict[self.domain_label] == i] = self.bns[i.item() + 1](
                    h[xdict[self.domain_label] == i]
                )
        else:
            h = self.bns[0](h)
        return h

    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()

    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()


def create_activation(name):
    if name == "relu":
        return nn.ReLU()
    elif name == "gelu":
        return nn.GELU()
    elif name == "prelu":
        return nn.PReLU()
    elif name is None:
        return nn.Identity()
    elif name == "elu":
        return nn.ELU()
    else:
        raise NotImplementedError(f"{name} is not implemented.")


def create_norm(name, n, h=4):
    if name == "layernorm":
        return nn.LayerNorm(n)
    elif name == "batchnorm":
        return nn.BatchNorm1d(n)
    elif name == "groupnorm":
        return nn.GroupNorm(h, n)
    elif name == "instancenorm":
        return nn.instancenorm1d(n)
    elif name == "rmsnorm":
        return RMSNorm(n)
    else:
        # print('Encoder without norm!')
        return nn.Identity()



import pandas as pd
import numpy as np
from sklearn.preprocessing import MinMaxScaler, StandardScaler


class NoScaler:
    def __init__(self):
        pass

    def fit(self, dfx):
        return self

    def transform(self, dfx):
        return dfx

    def fit_transform(self, dfx):
        return dfx

    def inverse_transform(self, dfx):
        return dfx


class P2Normalizer:
    def __init__(self, eps=1e-12):
        self.eps = eps

    def _square(self, x):
        return x**2

    def _clip(self, x):
        if x < self.eps:
            x = self.eps
        return x

    def fit(self, dfx):
        p2 = np.sqrt(dfx.apply(self._square, axis=1).sum())
        self.p2 = p2.apply(self._clip)
        dfx_normed = dfx / self.p2
        return dfx_normed

    def transform(self, dfx):
        df_scaled = dfx / self.p2
        return df_scaled


class Datascaler:

    def __init__(self, scale_method="minmax"):
        self.scale_method = scale_method
        if scale_method == "minmax":
            scaler = MinMaxScaler()
        elif scale_method == "standard":
            scaler = StandardScaler()
        elif scale_method == "p1norm":
            scaler = P2Normalizer()
        else:
            scaler = NoScaler()
        self.scaler = scaler

    def fit(self, dfcx):

        ## do not fit the first column
        dfx = dfcx[dfcx.columns[1:]]
        dfc = dfcx[dfcx.columns[0]]

        df = np.log2(dfx + 1)
        self.scaler.fit(df)
        return self

    def transform(self, dfcx):

        ## do not transform the first column
        dfx = dfcx[dfcx.columns[1:]]
        dfc = dfcx[[dfcx.columns[0]]]

        df = np.log2(dfx + 1)
        X = self.scaler.transform(df)
        dfx_scaled = pd.DataFrame(X, columns=dfx.columns, index=dfx.index)
        df_scaled = dfc.join(dfx_scaled)

        return df_scaled

    def fit_transform(self, dfx):
        self = self.fit(dfx)
        df_scaled = self.transform(dfx)
        return df_scaled



"""
Personal COMPASS Response Map (CRMap) utilities
==============================================

This module provides end-to-end helpers to construct and render a patient-level
**Personal COMPASS Response Map (CRMap)** — a layered graph that links
expression (gene TPM) → gene score → geneset score → concept (celltype) score → output (NR/R).

Functions
---------
- `prepare_crmap_data`:
    Run COMPASS inference and assemble all data required to build a CRMap.
    Specifically, this helper:
      1) loads the trained COMPASS model and runs prediction on an input
         expression matrix (`dfcx`);
      2) extracts latent scores at the gene, geneset, and concept (celltype)
         layers (`dfgn`, `dfgs`, `dfct`);
      3) (optionally) z-scores the above matrices over patients to standardize
         scale;
      4) derives correlation-based edge weights for each CRMap layer:
         genetpm→gene, gene→geneset, geneset→celltype, celltype→output;
      5) returns the four edge tables together with the tensors needed
         downstream.  The return order is exactly what
         `personal_crmap_data` expects.

- `personal_crmap_data`:
    Build a *patient-specific*, long-form, layered edge list that attaches
    per-node values for a chosen patient. This filters to requested high-level
    concepts (celltypes) and optionally keeps only the top-K genes per geneset.
    The output is a single DataFrame concatenating edges from all CRMap layers
    with `source_value` / `target_value` columns filled for the patient.

- `draw_personal_crmap`:
    Render the CRMap for a single patient as a left-to-right layered graph with
    curved edges. Node color encodes node value (with separate colormaps for
    genetpm vs. other layers); edge color encodes layer-specific weights.
    Multiple layout/label parameters control spacing, curvature, thresholds,
    and styling.

Typical workflow
----------------
1. Call `prepare_crmap_data(dfcx, ...)` to obtain:
   `(celltype2output, geneset2celltype, gene2geneset, genetpm2gene,
     dfgn, dfgs, dfct, dfpred, dfcx)`.
2. Call `personal_crmap_data(patient_id=..., concept2plot=..., TopK_gene=..., 
   celltype2output=..., geneset2celltype=..., gene2geneset=..., genetpm2gene=..., 
   dfgn=..., dfgs=..., dfct=..., dfpred=..., dfcx=...)` to build the
   patient-specific layered table.
3. Call `draw_personal_crmap(df, concept2plot=..., ...)` to visualize.

Inputs & assumptions
--------------------
- `dfcx` (patients × genes): input expression (e.g., TPM; columns are gene symbols,
  index are patient IDs).
- A trained COMPASS model loadable via `loadcompass(...)`, and projector/encoder
  weights accessible via `get_projector_weights(model)`.
- Latent extractions `model.extract(..., with_gene_level=True)` must return
  `(dfgn, dfgs, dfct)` aligned to the same patients as `dfcx`.
- Predictions `model.predict(...)` must yield a two-column `dfpred` that can be
  interpreted as `['NR','R']` (the helper normalizes column names if necessary).

Example
-------
>>> # 1) Prepare CRMap data from expression
>>> (celltype2output, geneset2celltype, gene2geneset, genetpm2gene,
...  dfgn, dfgs, dfct, dfpred, dfcx) = prepare_crmap_data(dfcx, z_scale=True)
>>>
>>> # 2) Build patient-specific layered table
>>> crmap_df = personal_crmap_data(
...     patient_id="P001",
...     concept2plot=["IFNg_pathway", "Cytotoxic_Tcell", 'Endothelial', 'TGFb_pathway'],
...     TopK_gene=3,
...     celltype2output=celltype2output,
...     geneset2celltype=geneset2celltype,
...     gene2geneset=gene2geneset,
...     genetpm2gene=genetpm2gene,
...     dfgn=dfgn, dfgs=dfgs, dfct=dfct, dfpred=dfpred, dfcx=dfcx
... )
>>>
>>> # 3) Draw the CRMap
>>> fig = draw_personal_crmap(crmap_df, concept2plot=["NKcell", "Reference"])
>>> fig.savefig("P001_CRMap.svg", bbox_inches="tight", dpi=150)

Dependencies
------------
- pandas, numpy, scikit-learn (for z-scoring, pairwise correlations),
- matplotlib (required), seaborn (optional for global style),
- project-specific utilities: `loadcompass`, `get_projector_weights`.

Notes
-----
- `prepare_crmap_data` returns values in the exact order expected by
  `personal_crmap_data`:
  `(celltype2output, geneset2celltype, gene2geneset, genetpm2gene,
    dfgn, dfgs, dfct, dfpred, dfcx)`.
- For dense graphs, consider increasing thresholds in
  `draw_personal_crmap(..., label_threshold=..., genetpm_gene_edge_threshold=...)`
  or reducing `TopK_gene` in `personal_crmap_data`.
"""

import warnings
from typing import Tuple, Optional, Dict
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import pairwise_distances
from .sankey import get_projector_weights
from .loader import loadcompass


def prepare_crmap_data(
    dfcx: pd.DataFrame,
    model_path_name: str = "./finetuner_pft_all.pt",
    map_location: str = "cpu",
    z_scale: bool = True,
    concept_palette: Optional[Dict[str, str]] = None,
) -> Tuple[
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
    pd.DataFrame,
]:
    """
    Run COMPASS inference and construct CRMap edge tables.

    Parameters
    ----------
    dfcx : DataFrame (patients x genes)
        Input expression (e.g., TPM; columns = gene symbols; index = patient IDs).
    model_path_name : str, default "finetuner_pft_all.pt"
        Path to the fine-tuned COMPASS checkpoint.
    map_location : str, default "cpu"
        Torch map_location (e.g., "cpu", "cuda:0").
    z_scale : bool, default True
        Z-score each of dfcx/dfgn/dfgs/dfct (column-wise over patients) before
        computing correlations and return the scaled versions.
    concept_palette : dict[str,str], optional
        Optional color mapping for concepts (celltypes) in celltype2output (key: concept name).

    Returns
    -------
    celltype2output : DataFrame
    geneset2celltype : DataFrame
    gene2geneset : DataFrame
    genetpm2gene : DataFrame
    dfgn : DataFrame
    dfgs : DataFrame
    dfct : DataFrame
    dfpred : DataFrame
    dfcx : DataFrame

    Notes
    -----
    - The four returned edge tables follow CRMap layers:
      genetpm→gene, gene→geneset, geneset→celltype, celltype→output.
    - `dfpred` keeps the model's original two columns (often ['$P_{NR}$','$P_{R}$']).
      Internally, we temporarily map them to ['NR','R'] to compute celltype→output correlations.
    """
    # -------- 1) Inference --------
    model = loadcompass(model_path_name, map_location=map_location)
    _, dfpred = model.predict(dfcx, batch_size=128)
    dfgn, dfgs, dfct = model.extract(dfcx, batch_size=128, with_gene_level=True)

    # Keep original dfpred for return; internally map to ['NR','R'] for correlation
    dfpred_out = dfpred.copy()
    dfp = dfpred_out.copy()
    if set(dfp.columns) == {"$P_{NR}$", "$P_{R}$"}:
        dfp.columns = ["NR", "R"]
    elif set(dfp.columns) != {"NR", "R"}:
        # minimal fallback: take first two columns as NR/R
        dfp = dfp.iloc[:, :2].copy()
        dfp.columns = ["NR", "R"]

    dfw = get_projector_weights(model).copy()
    dfw = dfw[["source", "target", "weights", "group", "concept"]]

    # -------- 2) Optional z-scoring (column-wise) --------
    def _z(df: pd.DataFrame) -> pd.DataFrame:
        scaler = StandardScaler().set_output(transform="pandas")
        return scaler.fit_transform(df)

    if z_scale:
        dfcx_z = _z(dfcx)
        dfgn_z = _z(dfgn)
        dfgs_z = _z(dfgs)
        dfct_z = _z(dfct)
        dfp_z = _z(dfp)  # only used for celltype→output correlation
    else:
        dfcx_z, dfgn_z, dfgs_z, dfct_z, dfp_z = dfcx, dfgn, dfgs, dfct, dfp

    # -------- 3) Align patients across all tables --------
    common_idx = dfct_z.index
    for m in (dfgs_z, dfgn_z, dfp_z, dfcx_z):
        common_idx = common_idx.intersection(m.index)
    if len(common_idx) == 0:
        raise ValueError("No overlapping patients across dfct/dfgs/dfgn/dfpred/dfcx.")

    dfct_ = dfct_z.loc[common_idx]
    dfgs_ = dfgs_z.loc[common_idx]
    dfgn_ = dfgn_z.loc[common_idx]
    dfcx_ = dfcx_z.loc[common_idx]
    dfp_ = dfp_z.loc[common_idx]

    # -------- 4) Build CRMap edge tables --------
    # celltype -> output
    corr_ct_out = 1 - pairwise_distances(dfct_.T, dfp_.T, metric="correlation")
    celltype2output = (
        pd.DataFrame(corr_ct_out, index=dfct_.columns, columns=["NR", "R"])
        .sort_values("R", ascending=False)
        .unstack()
        .reset_index()
    )
    celltype2output.columns = ["target", "source", "weights"]  # target in {'NR','R'}
    celltype2output["group"] = "celltype->output"
    celltype2output["concept"] = celltype2output["source"]
    celltype2output["source_color"] = (
        celltype2output["source"].map(concept_palette)
        if concept_palette is not None
        else "#9e9d93"
    )
    celltype2output["target_color"] = "#9e9d93"

    # geneset -> celltype
    geneset2celltype = dfw[dfw.group == "geneset->celltype"][
        ["source", "target", "group", "concept"]
    ].copy()
    corr_gs_ct = 1 - pairwise_distances(dfgs_.T, dfct_.T, metric="correlation")
    tmp = (
        pd.DataFrame(np.abs(corr_gs_ct), index=dfgs_.columns, columns=dfct_.columns)
        .T.unstack()
        .loc[geneset2celltype.set_index(["source", "target"]).index]
        .reset_index()
    )
    tmp.columns = ["source", "target", "weights"]
    geneset2celltype["weights"] = tmp["weights"].values

    # gene -> geneset
    gene2geneset = dfw[dfw.group == "gene->geneset"][
        ["source", "target", "group", "concept"]
    ].copy()
    corr_gn_gs = 1 - pairwise_distances(dfgn_.T, dfgs_.T, metric="correlation")
    tmp = (
        pd.DataFrame(np.abs(corr_gn_gs), index=dfgn_.columns, columns=dfgs_.columns)
        .T.unstack()
        .loc[gene2geneset.set_index(["source", "target"]).index]
        .reset_index()
    )
    tmp.columns = ["source", "target", "weights"]
    gene2geneset["weights"] = tmp["weights"].values

    # genetpm -> gene
    used_genes = [
        g
        for g in gene2geneset["source"].unique()
        if g in dfcx_.columns and g in dfgn_.columns
    ]
    if len(used_genes) == 0:
        raise ValueError(
            "No overlap between genes in dfcx/dfgn and 'gene->geneset' edges."
        )
    X = dfcx_[used_genes].values  # TPM
    Y = dfgn_[used_genes].values  # gene scores
    corr_cx_gn = 1 - pairwise_distances(X.T, Y.T, metric="correlation")
    attn = np.abs(corr_cx_gn)
    genetpm2gene = (
        pd.DataFrame(attn, columns=used_genes, index=used_genes).unstack().reset_index()
    )
    genetpm2gene.columns = ["source", "target", "weights"]
    genetpm2gene["source"] = genetpm2gene["source"] + "@TPM"
    genetpm2gene["group"] = "genetpm->gene"
    genetpm2gene["concept"] = "Gene"
    genetpm2gene["source_color"] = "#9e9d93"
    genetpm2gene["target_color"] = "#9e9d93"

    # -------- 5) Return in the exact order personal_crmap_data expects --------
    # (celltype2output, geneset2celltype, gene2geneset, genetpm2gene, dfgn, dfgs, dfct, dfpred, dfcx)
    return (
        celltype2output,
        geneset2celltype,
        gene2geneset,
        genetpm2gene,
        dfgn_,
        dfgs_,
        dfct_,
        dfpred_out,
        dfcx_,
    )


def personal_crmap_data(
    patient_id,
    concept2plot=["NKcell", "Reference"],
    TopK_gene=1,
    celltype2output=None,
    geneset2celltype=None,
    gene2geneset=None,
    genetpm2gene=None,
    dfgn=None,
    dfgs=None,
    dfct=None,
    dfpred=None,
    dfcx=None,
):
    """
    Build the patient-specific layered edge table used by `draw_personal_crmap`.

    Given a patient identifier and the projector/encoder tables that link layers
    (genetpm→gene→geneset→celltype→output), this function assembles a long-form
    DataFrame of edges with associated node values for the chosen patient. It
    also restricts the visualization to a subset of concepts (celltypes) and,
    optionally, the top-k contributing genes per geneset.

    Parameters
    ----------
    patient_id : hashable
        Identifier that exists as an index in all node-value frames (`dfgn`,
        `dfgs`, `dfct`, `dfpred`, `dfcx`).
    concept2plot : list of str, default ['NKcell', 'Reference']
        Celltype (high-level concepts) to include in the final graph.
    TopK_gene : int, default 1
        For each geneset, keep the top-k genes by absolute `weights` in the
        `gene2geneset` table.
    celltype2output : pandas.DataFrame
        Edge list mapping celltype → output. Required columns:
        ['source', 'target', 'weights']; `source` is a celltype; `target` in {'R', 'NR'}.
    geneset2celltype : pandas.DataFrame
        Edge list mapping geneset → celltype. Required columns:
        ['source', 'target', 'weights'].
    gene2geneset : pandas.DataFrame
        Edge list mapping gene → geneset. Required columns:
        ['source', 'target', 'weights'].
    genetpm2gene : pandas.DataFrame
        Edge list mapping (gene@TPM) → gene. Required columns:
        ['source', 'target', 'weights'] where `source` uses the convention
        '<gene>@TPM'.
    dfgn : pandas.DataFrame
        Gene-level node values (e.g., z-scored gene scores). Index are patient
        IDs; columns are gene symbols.
    dfgs : pandas.DataFrame
        Geneset-level node values. Index are patient IDs; columns are genesets.
    dfct : pandas.DataFrame
        Celltype (concept)–level node values. Index are patient IDs; columns are
        celltypes in your ontology.
    dfpred : pandas.DataFrame
        Output probabilities/scores. Index are patient IDs; columns must contain
        two entries that will be assigned to ['NR', 'R'] for visualization.
    dfcx : pandas.DataFrame
        Gene TPM–level node values (input). Index are patient IDs; columns are
        gene symbols (not including '@TPM'); values typically z-scored TPM.

    Returns
    -------
    pandas.DataFrame
        Long-form edge table concatenating four groups:
        - 'celltype->output'
        - 'geneset->celltype'
        - 'gene->geneset'
        - 'genetpm->gene'
        With the following columns:
        ['source', 'target', 'weights', 'group', 'source_value', 'target_value'].

    Raises
    ------
    ValueError
        If any required frame is None, if `patient_id` is not found, or if
        required columns are missing.

    Notes
    -----
    - `dfpred.loc[patient]` is reindexed to ['NR', 'R'] to standardize output ordering.
    - Genes are filtered to `TopK_gene` per geneset by descending 'weights'.
    - For `genetpm2gene`, `source` is matched after stripping '@TPM' to align
      with `dfcx` (gene TPM values).
    """
    # ---- original function body of personal_camap_data goes here unchanged ----
    # Simply paste your existing implementation body here.
    # Only the function name & docstring changed.
    if any(
        v is None
        for v in [
            celltype2output,
            geneset2celltype,
            gene2geneset,
            genetpm2gene,
            dfgn,
            dfgs,
            dfct,
            dfpred,
            dfcx,
        ]
    ):
        raise ValueError("Please provide all required data.")
    try:
        _ = dfcx.loc[patient_id]
        patient = patient_id
    except KeyError:
        raise ValueError(f"Patient ID {patient_id} not found in clinical data.")

    dfp1 = celltype2output[celltype2output["source"].isin(concept2plot)].reset_index(
        drop=True
    )
    dfp1["source_value"] = dfp1.source.map(dfct.loc[patient])
    pred = dfpred.loc[patient]
    pred.index = ["NR", "R"]
    dfp1["target_value"] = dfp1.target.map(pred)

    dfp2 = geneset2celltype[geneset2celltype.target.isin(concept2plot)].reset_index(
        drop=True
    )
    dfp2["source_value"] = dfp2.source.map(dfgs.loc[patient])
    dfp2["target_value"] = dfp2.target.map(dfct.loc[patient])

    dfp3 = gene2geneset[gene2geneset.target.isin(dfp2.source)]
    dfp3 = (
        dfp3.groupby("target")
        .apply(lambda x: x.sort_values("weights", ascending=False).iloc[:TopK_gene])
        .reset_index(drop=True)
    )
    dfp3["source_value"] = dfp3.source.map(dfgn.loc[patient])
    dfp3["target_value"] = dfp3.target.map(dfgs.loc[patient])

    dfp4 = genetpm2gene[
        genetpm2gene.source.apply(lambda x: x.split("@")[0]).isin(dfp3.source.unique())
    ]
    dfp4 = dfp4[dfp4.target.isin(dfp3.source.unique())]
    dfp4["source_value"] = dfp4.source.apply(lambda x: x.split("@")[0]).map(
        dfcx.loc[patient]
    )
    dfp4["target_value"] = dfp4.target.map(dfgn.loc[patient])

    data_r = pd.concat([dfp1, dfp2, dfp3, dfp4]).reset_index(drop=True)
    return data_r


import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.patches import FancyArrowPatch
from matplotlib.colors import Normalize, LinearSegmentedColormap
from matplotlib.cm import get_cmap, ScalarMappable
import numpy as np
import seaborn as sns
from itertools import chain
import io


def draw_personal_crmap(
    df,
    concept2plot,
    figsize=(16, 15),
    fontsize=15,
    layer_node_sizes=[0.5, 0.5, 1, 3, 5],
    max_rad=0.25,
    show_geneset_name=False,
    layer_node_gaps={"celltype": 0.15, "geneset": 0.1, "output": 0.2},
    layer_spacing=[1.2, 1, 1, 0.8],
    node_label_alignments={
        "genetpm": -0.01,
        "gene": -0.01,
        "geneset": 0.01,
        "celltype": 0.02,
        "output": 0.02,
    },
    layer_edge_styles={"celltype->output": {"linestyle": "--", "arrowstyle": "-"}},
    genetpm_node_color_map={"cmap": None, "vmin": -2, "vmax": 2},
    compass_node_color_map={"cmap": None, "vmin": 0, "vmax": 1},
    label_threshold={
        "genetpm": 1,
        "gene": 0.5,
        "geneset": 0.5,
        "celltype": -10,
        "output": -10,
    },
    label_color_same_as_node=True,
    genetpm2gene_edge_colormap={"cmap": None, "vmin": 0.6, "vmax": 1.0},
    projector_edge_colormap={"cmap": None, "vmin": 0.0, "vmax": 0.5},
    edge_line_width=[0.2, 0.3, 0.3, 0.5],
    genetpm_gene_edge_threshold=0.6,
    edge_activate={"threshold": 10, "color": "red"},
):
    """
    Render a multi-layer, curved-edge Personal COMPASS Response Map (CRMap).

    This function draws a left-to-right layered graph with nodes arranged by
    layer in the order:
        genetpm → gene → geneset → celltype → output
    Node colors reflect node values (two independent colormaps for genetpm vs
    the rest), and edges are colored by their layer-specific colormap. Several
    layout and styling knobs are exposed to control spacing, label density, and
    curvature.

    Parameters
    ----------
    df : pandas.DataFrame
        The patient-specific layered edge table returned by `personal_crmap_data`.
        Required columns:
        - 'source', 'target', 'weights'
        - 'group' with values in {'genetpm->gene','gene->geneset','geneset->celltype','celltype->output'}
        - 'source_value', 'target_value'
    concept2plot : list of str
        High-level concepts (celltypes) to prioritize in ordering (right side).
        If present, they are placed in reverse order for vertical alignment.
    figsize : tuple, default (16, 15)
        Matplotlib figure size (inches).
    fontsize : int, default 15
        Base font size for labels and colorbar titles.
    layer_node_sizes : list of float, default [0.5, 0.5, 1, 3, 5]
        Relative node sizes per layer (genetpm→…→output). Values are scaled by 100 internally.
    max_rad : float, default 0.25
        Maximum edge curvature (arc radius) in `FancyArrowPatch` connection styles.
    show_geneset_name : bool, default False
        Whether to draw geneset labels (can be dense).
    layer_node_gaps : dict, optional
        Vertical spacing per layer (fraction of axis height between neighboring nodes).
    layer_spacing : list of float, default [1.2, 1, 1, 0.8]
        Horizontal spacing between consecutive layers (linearly normalized to [0,1]).
    node_label_alignments : dict, optional
        Per-layer horizontal text offsets; negative values left-align, positive
        right-align relative to node center.
    layer_edge_styles : dict, optional
        Mapping from '<source_layer->target_layer>' to style dicts with keys
        'linestyle' and 'arrowstyle'.
    genetpm_node_color_map : dict, optional
        Node colormap spec for genetpm layer: {'cmap': name or None, 'vmin': float, 'vmax': float}.
        If `cmap` is None, a default blue–white–red map is used.
    compass_node_color_map : dict, optional
        Node colormap spec for gene/geneset/celltype/output layers. If `cmap`
        is None, a default white→yellow→orange→red map is used.
    label_threshold : dict, optional
        Minimum node value to show a text label per layer. Set very low (e.g.,
        -10) to always show; increase to sparsify labels.
    label_color_same_as_node : bool, default True
        If True, gene/TPM labels adopt the node color, improving saliency.
    genetpm2gene_edge_colormap : dict, optional
        Colormap spec for genetpm→gene edges.
    projector_edge_colormap : dict, optional
        Colormap spec for projector edges (gene→geneset→celltype→output).
    edge_line_width : list of float, default [0.2, 0.3, 0.3, 0.5]
        Line widths per edge group following the layer order above.
    genetpm_gene_edge_threshold : float, default 0.6
        Prunes genetpm→gene edges with |weight| below this threshold (except
        the self-named “gene@TPM → gene” identity edges).
    edge_activate : dict, optional
        If provided as {'threshold': float, 'color': str}, re-colors edges with
        weight > threshold to the given color.

    Returns
    -------
    matplotlib.figure.Figure
        The Matplotlib figure (axes are created internally).

    Notes
    -----
    - Layer order is inferred from `df['group']` and fixed to
      genetpm→gene→geneset→celltype→output.
    - The function builds node positions with per-layer vertical spacing controls
      (`layer_node_gaps`) and per-layer horizontal spacing (`layer_spacing`).
    - Four colorbars are added (genetpm nodes, compass nodes, encoder edges, projector edges).
    - To save the figure, call `fig.savefig(...)` on the returned `Figure`.

    Examples
    --------
    >>> camap = personal_camap_data(
    ...     patient_id="P001",
    ...     concept2plot=["CD8_T", "NKcell"],
    ...     TopK_gene=3,
    ...     celltype2output=celltype2output,
    ...     geneset2celltype=geneset2celltype,
    ...     gene2geneset=gene2geneset,
    ...     genetpm2gene=genetpm2gene,
    ...     dfgn=dfgn, dfgs=dfgs, dfct=dfct, dfpred=dfpred, dfcx=dfcx,
    ... )
    >>> fig = draw_personal_camap(camap, concept2plot=["CD8_T", "NKcell"], figsize=(14, 12))
    >>> fig.savefig("P001_camap.svg", bbox_inches="tight", dpi=150)

    """
    # ---- original function body of draw_personal_camap goes here unchanged ----
    # Paste your existing implementation body here; only the name/docstring changed.
    fig, ax = plt.subplots(figsize=figsize)
    ax.axis("off")

    layer_links_order = [
        "genetpm->gene",
        "gene->geneset",
        "geneset->celltype",
        "celltype->output",
    ]
    df["weights_abs"] = df["weights"].abs()
    layer_node_order_series = df.groupby(["group", "target"]).apply(
        lambda x: x.sort_values("weights", ascending=True).source.tolist()
    )
    layer_node_order = {}
    try:
        layer_node_order["celltype"] = concept2plot[
            ::-1
        ]  # layer_node_order_series.loc[('celltype->output', 'R')]
    except KeyError:
        layer_node_order["celltype"] = []
    try:
        geneset_order_series = (
            df[df.group == "geneset->celltype"]
            .groupby("target")
            .apply(lambda x: x.sort_values("weights").source.tolist())
            .loc[layer_node_order["celltype"]]
        )
        geneset_order = list(chain.from_iterable(geneset_order_series))
        layer_node_order["geneset"] = geneset_order
    except KeyError:
        layer_node_order["geneset"] = []
    layer_node_order["output"] = ["NR", "R"]

    if "group" not in df.columns:
        raise ValueError("DataFrame must contain a 'group' column.")
    geneset_celltype_mapping = (
        df[df.group == "geneset->celltype"].set_index("source")["target"].to_dict()
    )

    if genetpm_node_color_map.get("cmap") is None:
        genetpm_colors = ["blue", "white", "red"]
        cmap_genetpm = LinearSegmentedColormap.from_list(
            "genetpm_default", genetpm_colors
        )
    else:
        cmap_genetpm = plt.get_cmap(genetpm_node_color_map["cmap"])
    vmin_genetpm = genetpm_node_color_map.get("vmin", None)
    vmax_genetpm = genetpm_node_color_map.get("vmax", None)

    if compass_node_color_map.get("cmap") is None:
        compass_colors = ["white", "yellow", "orange", "red"]
        cmap_compass = LinearSegmentedColormap.from_list(
            "compass_default", compass_colors
        )
    else:
        cmap_compass = plt.get_cmap(compass_node_color_map["cmap"])
    vmin_compass = compass_node_color_map.get("vmin", None)
    vmax_compass = compass_node_color_map.get("vmax", None)

    if genetpm2gene_edge_colormap.get("cmap") is None:
        g2g_colors = [
            "#c6dbef",
            "#9ecae1",
            "#6baed6",
            "#4292c6",
            "#2171b5",
            "#08519c",
            "#08306b",
            "#041f47",
        ]
        g2g_cmap = LinearSegmentedColormap.from_list(
            "genetpm2gene_edge_default", g2g_colors
        )
    else:
        g2g_cmap = plt.get_cmap(genetpm2gene_edge_colormap["cmap"])
    g2g_vmin = genetpm2gene_edge_colormap["vmin"]
    g2g_vmax = genetpm2gene_edge_colormap["vmax"]
    g2g_norm = Normalize(vmin=g2g_vmin, vmax=g2g_vmax)

    if projector_edge_colormap.get("cmap") is None:
        proj_colors = [
            "#C7E9B4",
            "#7FCDBB",
            "#41B6C4",
            "#1D91C0",
            "#225EA8",
            "#253494",
            "#081D58",
            "#08306b",
            "#041F47",
        ]
        proj_cmap = LinearSegmentedColormap.from_list(
            "projector_edge_default", proj_colors
        )
    else:
        proj_cmap = plt.get_cmap(projector_edge_colormap["cmap"])
    proj_vmin = projector_edge_colormap["vmin"]
    proj_vmax = projector_edge_colormap["vmax"]
    proj_norm = Normalize(vmin=proj_vmin, vmax=proj_vmax)

    df[["source_layer", "target_layer"]] = df["group"].str.split("->", expand=True)

    all_layers = []
    for link in layer_links_order:
        layers = link.split("->")
        for layer in layers:
            if layer not in all_layers:
                all_layers.append(layer)
    n_layers = len(all_layers)

    if "source_value" not in df.columns or "target_value" not in df.columns:
        raise ValueError(
            "DataFrame must contain 'source_value' and 'target_value' columns."
        )
    node_values = {}
    node_layers = {}
    for idx, rowd in df.iterrows():
        node_values[rowd["source"]] = rowd["source_value"]
        node_values[rowd["target"]] = rowd["target_value"]
        node_layers[rowd["source"]] = rowd["source_layer"]
        node_layers[rowd["target"]] = rowd["target_layer"]

    layer_nodes_dict = {}
    for node, layer in node_layers.items():
        layer_nodes_dict.setdefault(layer, set()).add(node)

    layer_node_sizes = [size * 100 for size in layer_node_sizes]

    for layer in all_layers:
        nodes = layer_nodes_dict.get(layer, set())
        if layer in layer_node_order:
            ordered_nodes = [node for node in layer_node_order[layer] if node in nodes]
            unordered_nodes = [node for node in nodes if node not in ordered_nodes]
            layer_nodes = ordered_nodes + unordered_nodes
        else:
            if layer == "geneset" and geneset_celltype_mapping:
                geneset_groups = {}
                for node in nodes:
                    celltype = geneset_celltype_mapping.get(node, "Unknown")
                    geneset_groups.setdefault(celltype, []).append(node)
                layer_nodes = []
                for celltype in sorted(geneset_groups.keys()):
                    group_nodes = sorted(geneset_groups[celltype], reverse=True)
                    layer_nodes.extend(group_nodes)
            else:
                layer_nodes = sorted(nodes, reverse=True)
        layer_nodes_dict[layer] = layer_nodes

    genetpm_layer = "genetpm"
    genetpm_nodes = layer_nodes_dict.get(genetpm_layer, [])
    compass_nodes = []
    for layer in all_layers:
        if layer != genetpm_layer:
            compass_nodes.extend(layer_nodes_dict.get(layer, []))

    genetpm_values = [node_values[n] for n in genetpm_nodes] if genetpm_nodes else []
    compass_values = [node_values[n] for n in compass_nodes] if compass_nodes else []

    if vmin_genetpm is None and genetpm_values:
        vmin_genetpm = min(genetpm_values)
    if vmax_genetpm is None and genetpm_values:
        vmax_genetpm = max(genetpm_values)
    if vmin_compass is None and compass_values:
        vmin_compass = min(compass_values)
    if vmax_compass is None and compass_values:
        vmax_compass = max(compass_values)

    norm_genetpm = Normalize(vmin=vmin_genetpm, vmax=vmax_genetpm)
    norm_compass = Normalize(vmin=vmin_compass, vmax=vmax_compass)

    # Use layer_spacing to determine layer positions
    if len(layer_spacing) < n_layers - 1:
        layer_spacing = list(layer_spacing) + [1] * (n_layers - 1 - len(layer_spacing))
    layer_spacing = layer_spacing[: n_layers - 1]

    x_positions = [0]
    for sp in layer_spacing:
        x_positions.append(x_positions[-1] + sp)
    max_pos = x_positions[-1] if x_positions else 1
    if max_pos == 0:
        max_pos = 1
    x_positions = [x / max_pos for x in x_positions]

    layer_positions = {layer: x_positions[idx] for idx, layer in enumerate(all_layers)}

    node_positions = {}
    for idx, layer_name in enumerate(all_layers):
        nodes = layer_nodes_dict.get(layer_name, [])
        n_nodes = len(nodes)
        x_position = layer_positions[layer_name]
        node_size = layer_node_sizes[idx]

        if n_nodes > 1:
            if layer_name in layer_node_gaps:
                gap_y = layer_node_gaps[layer_name]
                total_height = gap_y * (n_nodes - 1)
                if total_height > 1:
                    gap_y = 1.0 / (n_nodes - 1)
                y_start = 0.5 - (gap_y * (n_nodes - 1) / 2)
                y_positions = [y_start + i * gap_y for i in range(n_nodes)]
            else:
                y_positions = np.linspace(0, 1, n_nodes)
        else:
            y_positions = [0.5]

        y_positions = [min(max(y, 0), 1) for y in y_positions]
        positions = list(zip([x_position] * n_nodes, y_positions))

        if layer_name == "genetpm":
            c_map = cmap_genetpm
            c_norm = norm_genetpm
        else:
            c_map = cmap_compass
            c_norm = norm_compass

        x_positions_plot = [pos[0] for pos in positions]
        y_positions_plot = [pos[1] for pos in positions]
        node_colors = [c_map(c_norm(node_values[n])) for n in nodes]
        # node_colors = [mcolors.to_hex(c_map(c_norm(node_values[n]))) for n in nodes]

        ax.scatter(
            x_positions_plot,
            y_positions_plot,
            s=node_size,
            zorder=3,
            linewidths=0.5,
            edgecolors="k",
            facecolors=node_colors,
        )
        for n, (xx, yy) in zip(nodes, positions):
            node_positions[n] = (xx, yy)

        if label_threshold and layer_name in label_threshold:
            threshold = label_threshold[layer_name]
            offset = node_label_alignments.get(layer_name, -0.02)
            i = 0
            for n, (xx, yy) in zip(nodes, positions):
                if layer_name == "geneset" and not show_geneset_name:
                    continue

                if layer_name == "genetpm":
                    node_value = abs(node_values[n])
                else:
                    node_value = node_values[n]

                if node_value > threshold:
                    if offset < 0:
                        text_x = xx + offset
                        ha = "right"
                    elif offset > 0:
                        text_x = xx + offset
                        ha = "left"
                    else:
                        text_x = xx
                        ha = "center"
                    label_text = str(n).replace("@TPM", "")

                    if layer_name == "output":
                        label_text = "$P_{%s} = %s$" % (
                            label_text,
                            round(node_value, 2),
                        )

                    if label_color_same_as_node and (
                        layer_name == "genetpm"
                    ):  # or layer_name == 'gene'
                        ax.text(
                            text_x,
                            yy,
                            label_text,
                            fontsize=fontsize,
                            ha=ha,
                            va="center",
                            color=node_colors[i],
                        )
                    else:
                        ax.text(
                            text_x,
                            yy,
                            label_text,
                            fontsize=fontsize,
                            ha=ha,
                            va="center",
                        )
                i += 1

        layer_name_map = {
            "genetpm": "INPUT" + "\n$X_{GeneTPM}$",
            "gene": "Gene" + "\nscore: $S_{Gene}$",
            "geneset": "Granular concept" + "\nscore: $S_{Geneset}$",
            "celltype": "High-level concept" + "\nscore: $S_{Concept}$",
            "output": "OUTPUT" + "\n$P_{(R|NR)}$",
        }

        if layer_name != "output":
            ax.text(
                x_position,
                1.07,
                layer_name_map.get(layer_name),
                fontsize=fontsize + 2,
                ha="center",
                va="center",
            )
        else:
            ax.text(
                x_position,
                0.5,
                layer_name_map.get(layer_name),
                fontsize=fontsize + 2,
                ha="center",
                va="center",
            )

    for idx, rowd in df.iterrows():
        source = rowd["source"]
        target = rowd["target"]
        weight = rowd["weights"]
        source_layer = rowd["source_layer"]
        target_layer = rowd["target_layer"]
        layer_connection = f"{source_layer}->{target_layer}"

        # threshold pruning for genetpm->gene
        if (
            layer_connection == "genetpm->gene"
            and abs(weight) < genetpm_gene_edge_threshold
            and (str(source).replace("@TPM", "") != target)
        ):
            continue

        if layer_connection == "genetpm->gene":
            edge_norm_value = g2g_norm(weight)
            current_edge_color = g2g_cmap(edge_norm_value)
            lw = edge_line_width[layer_links_order.index(layer_connection)]
        else:
            edge_norm_value = proj_norm(weight)
            current_edge_color = proj_cmap(edge_norm_value)
            lw = edge_line_width[layer_links_order.index(layer_connection)]

        if edge_activate:
            threshold = edge_activate.get("threshold", None)
            override_color = edge_activate.get("color", None)
            if threshold is not None and override_color is not None:
                if weight > threshold:
                    current_edge_color = override_color

        (x_a, y_a) = node_positions[source]
        (x_b, y_b) = node_positions[target]
        delta_y = y_b - y_a
        rad = max_rad * delta_y
        rad = max(-max_rad, min(max_rad, rad))

        edge_style = layer_edge_styles.get(
            layer_connection, {"linestyle": "-", "arrowstyle": "-"}
        )
        line_style = edge_style.get("linestyle", "-")
        arrow_style = edge_style.get("arrowstyle", "-")

        con = FancyArrowPatch(
            (x_a, y_a),
            (x_b, y_b),
            arrowstyle=arrow_style,
            linestyle=line_style,
            connectionstyle=f"arc3,rad={rad}",
            color=current_edge_color,
            linewidth=lw,
            mutation_scale=10,
        )
        ax.add_patch(con)

    fig.subplots_adjust(right=0.88)

    sm_genetpm = ScalarMappable(cmap=cmap_genetpm, norm=norm_genetpm)
    sm_genetpm.set_array([])
    cax_genetpm = fig.add_axes([0.94, 0.74, 0.015, 0.1])
    cbar_genetpm = plt.colorbar(sm_genetpm, cax=cax_genetpm, orientation="vertical")
    cbar_genetpm.set_label(
        "Gene TPM\nZ-score", rotation=90, labelpad=5, fontsize=fontsize
    )
    cbar_genetpm.ax.tick_params(labelsize=fontsize)
    cbar_genetpm.set_ticks([vmin_genetpm, vmax_genetpm])
    cbar_genetpm.set_ticklabels([f"{vmin_genetpm:.1f}", f"{vmax_genetpm:.1f}"])

    sm_compass = ScalarMappable(cmap=cmap_compass, norm=norm_compass)
    sm_compass.set_array([])
    cax_compass = fig.add_axes([0.94, 0.56, 0.015, 0.1])
    cbar_compass = plt.colorbar(sm_compass, cax=cax_compass, orientation="vertical")
    cbar_compass.set_label(
        "COMPASS score\nZ-score", rotation=90, labelpad=5, fontsize=fontsize
    )
    cbar_compass.ax.tick_params(labelsize=fontsize)
    cbar_compass.set_ticks([vmin_compass, vmax_compass])
    cbar_compass.set_ticklabels([f"{vmin_compass:.1f}", f"{vmax_compass:.1f}"])

    sm_g2g = ScalarMappable(cmap=g2g_cmap, norm=g2g_norm)
    sm_g2g.set_array([])
    cax_g2g = fig.add_axes([0.94, 0.37, 0.015, 0.1])
    cbar_g2g = plt.colorbar(sm_g2g, cax=cax_g2g, orientation="vertical")
    cbar_g2g.set_label(
        "Encoder\nedge weights", rotation=90, labelpad=5, fontsize=fontsize
    )
    cbar_g2g.ax.tick_params(labelsize=fontsize)
    cbar_g2g.set_ticks([g2g_vmin, g2g_vmax])
    cbar_g2g.set_ticklabels([f"{g2g_vmin:.1f}", f"{g2g_vmax:.1f}"])

    sm_proj = ScalarMappable(cmap=proj_cmap, norm=proj_norm)
    sm_proj.set_array([])
    cax_proj = fig.add_axes([0.94, 0.19, 0.015, 0.1])
    cbar_proj = plt.colorbar(sm_proj, cax=cax_proj, orientation="vertical")
    cbar_proj.set_label(
        "Projector\nedge weights", rotation=90, labelpad=5, fontsize=fontsize
    )
    cbar_proj.ax.tick_params(labelsize=fontsize)
    cbar_proj.set_ticks([proj_vmin, proj_vmax])
    cbar_proj.set_ticklabels([f"{proj_vmin:.1f}", f"{proj_vmax:.1f}"])

    plt.close(fig)
    return fig


if __name__ == "__main__":
    # Example
    # crmap_data = personal_crmap_data(
    #     patient_id=patient_id,
    #     concept2plot=concept2plot,
    #     TopK_gene=TopK_gene,
    #     celltype2output=celltype2output,
    #     geneset2celltype=geneset2celltype,
    #     gene2geneset=gene2geneset,
    #     genetpm2gene=genetpm2gene,
    #     dfgn=dfgn, dfgs=dfgs, dfct=dfct, dfpred=dfpred, dfcx=dfcx,
    # )
    # crmap_fig = draw_personal_crmap(crmap_data, concept2plot=concept2plot)
    # crmap_fig.savefig("CRMap_example.svg", bbox_inches="tight", dpi=150)
    pass



from torch._C import dtype

# !pip install torch
# !pip install einops

import math
from collections import namedtuple
from functools import wraps
from packaging import version

import torch
from torch import nn, einsum, Tensor
import torch.nn.functional as F

from einops import rearrange

from dataclasses import dataclass

# constants

EfficientAttentionConfig = namedtuple(
    "EfficientAttentionConfig", ["enable_flash", "enable_math", "enable_mem_efficient"]
)

# helpers


def exists(val):
    return val is not None


def once(fn):
    called = False

    @wraps(fn)
    def inner(x):
        nonlocal called
        if called:
            return
        called = True
        return fn(x)

    return inner


print_once = once(print)

# main class


@dataclass
class Intermediates:
    qk_similarities: Tensor = None
    pre_softmax_attn: Tensor = None
    post_softmax_attn: Tensor = None

    def to_tuple(self):
        return (self.qk_similarities, self.pre_softmax_attn, self.post_softmax_attn)


# helpers


class FlashAttention(nn.Module):
    def __init__(self, causal=False, dropout=0.0, flash=True):
        super().__init__()
        self.dropout = dropout
        self.attn_dropout = nn.Dropout(dropout)

        self.causal = causal
        self.flash = flash
        assert not (
            flash and version.parse(torch.__version__) < version.parse("2.0.0")
        ), "in order to use flash attention, you must be using pytorch 2.0 or above"

        # determine efficient attention configs for cuda and cpu

        self.cpu_config = EfficientAttentionConfig(True, True, True)
        self.cuda_config = None

        if not torch.cuda.is_available() or not flash:
            return

        device_properties = torch.cuda.get_device_properties(torch.device("cuda"))

        if device_properties.major == 8 and device_properties.minor == 0:
            print_once(
                "A100 GPU detected, using flash attention if input tensor is on cuda"
            )
            self.cuda_config = EfficientAttentionConfig(True, False, False)
        else:
            print_once(
                "Non-A100 GPU detected, using math or mem efficient attention if input tensor is on cuda"
            )
            self.cuda_config = EfficientAttentionConfig(False, True, True)

    def get_mask(self, i, j, device):
        return torch.ones((i, j), device=device, dtype=torch.bool).triu(j - i + 1)

    def flash_attn(self, q, k, v, mask=None, attn_bias=None):
        batch, heads, q_len, _, k_len, is_cuda, device = (
            *q.shape,
            k.shape[-2],
            q.is_cuda,
            q.device,
        )

        # Recommended for multi-query single-key-value attention by Tri Dao
        # kv shape torch.Size([1, 512, 64]) -> torch.Size([1, 8, 512, 64])

        if k.ndim == 3:
            k = rearrange(k, "b ... -> b 1 ...").expand_as(q)

        if v.ndim == 3:
            v = rearrange(v, "b ... -> b 1 ...").expand_as(q)

        # handle scale - by default they scale by dim_head ** -0.5, but need to take care if using cosine sim attention
        # Check if mask exists and expand to compatible shape
        # The mask is B L, so it would have to be expanded to B H N L

        causal = self.causal

        if exists(mask):
            assert mask.ndim == 4
            mask = mask.expand(batch, heads, q_len, k_len)

            # manually handle causal mask, if another mask was given

            if causal:
                causal_mask = self.create_causal_mask(q_len, k_len, device=device)
                mask = mask & ~causal_mask
                causal = False

        # handle alibi positional bias
        # convert from bool to float

        if exists(attn_bias):
            attn_bias = rearrange(attn_bias, "h i j -> 1 h i j").expand(
                batch, -1, -1, -1
            )

            # if mask given, the mask would already contain the causal mask from above logic
            # otherwise, if no mask given but still causal, mask out alibi positional bias to a large negative number

            mask_value = -torch.finfo(q.dtype).max

            if exists(mask):
                attn_bias = attn_bias.masked_fill(~mask, mask_value // 2)
            elif causal:
                causal_mask = self.create_causal_mask(q_len, k_len, device=device)
                attn_bias = attn_bias.masked_fill(causal_mask, mask_value // 2)
                causal = False

            # scaled_dot_product_attention handles attn_mask either as bool or additive bias
            # make it an additive bias here

            mask = attn_bias

        # Check if there is a compatible device for flash attention

        config = self.cuda_config if is_cuda else self.cpu_config

        # pytorch 2.0 flash attn: q, k, v, mask, dropout, causal, softmax_scale

        with torch.backends.cuda.sdp_kernel(**config._asdict()):
            out = F.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=mask,
                dropout_p=self.dropout if self.training else 0.0,
                is_causal=causal,
            )

            return out

    def forward(self, q, k, v, mask=None, attn_bias=None, need_weights=False):

        # print(q.shape, k.shape)

        """
        einstein notation
        b - batch
        h - heads
        n, i, j - sequence length (base sequence length, source, target)
        d - feature dimension
        """

        q_len, k_len, device = q.shape[-2], k.shape[-2], q.device

        scale = q.shape[-1] ** -0.5

        kv_einsum_eq = "b j d" if k.ndim == 3 else "b h j d"

        if self.flash:
            return self.flash_attn(q, k, v, mask=mask, attn_bias=attn_bias)

        # similarity
        sim = einsum(f"b h i d, {kv_einsum_eq} -> b h i j", q, k) * scale

        # attention bias
        if exists(attn_bias):
            sim = sim + attn_bias

        # causal mask
        if self.causal:
            causal_mask = self.get_mask(q_len, k_len, device)
            sim = sim.masked_fill(causal_mask, -torch.finfo(sim.dtype).max)

        # attention
        attn = sim.softmax(dim=-1)

        attn = self.attn_dropout(attn)

        # aggregate values
        out = einsum(f"b h i j, {kv_einsum_eq} -> b h i d", attn, v)

        if need_weights:
            return out, attn
        else:
            return out, None


class FlashMHA(nn.Module):
    def __init__(
        self,
        embed_dim,
        num_heads,
        bias=True,
        batch_first=True,
        dropout=0.0,
        causal=False,
        device=None,
        dtype=None,
        flash=True,
    ) -> None:
        assert batch_first
        factory_kwargs = {"device": device, "dtype": dtype}

        # print(factory_kwargs)

        super().__init__()
        self.embed_dim = embed_dim
        self.causal = causal

        self.num_heads = num_heads
        assert (
            self.embed_dim % num_heads == 0
        ), "self.kdim must be divisible by num_heads"
        self.head_dim = self.embed_dim // num_heads
        assert (
            self.head_dim % 8 == 0 and self.head_dim <= 128
        ), "Only support head_dim <= 128 and divisible by 8"

        self.Wqkv = nn.Linear(embed_dim, 3 * embed_dim, bias=bias, **factory_kwargs)
        self.inner_attn = FlashAttention(dropout=dropout, causal=causal, flash=flash)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias, **factory_kwargs)

    def forward(self, query, key, value, need_weights=False):

        qkv = self.Wqkv(query)

        q, k, v = rearrange(
            qkv,
            "b s (three h d) -> three b h s d",
            three=3,
            h=self.num_heads,
            d=self.head_dim,
        ).unbind(dim=0)

        # 128, 15674, 2, 16
        # input needs: b, h=2, s=15672, d=16

        print(q.shape, k.shape)  # should be batch, head, seq_len, dim

        context, attn = self.inner_attn(q, k, v, need_weights=need_weights)
        out = self.out_proj(rearrange(context, "b h s d -> b s (h d)"))
        return out, attn



import torch
import torch.nn as nn
import torch.nn.functional as F


class GeneSetPoolingAggregator(nn.Module):
    def __init__(self, genesets_indices, pooling_type="mean"):
        """
        Initializes the GeneSetPoolingAggregator module.

        :param genesets_indices: A list of lists, where each inner list contains indices of genes in a gene set.
        :param pooling_type: A string indicating the type of pooling ('mean' or 'max').
        """
        super(GeneSetPoolingAggregator, self).__init__()
        self.genesets_indices = genesets_indices
        self.pooling_type = pooling_type

    def forward(self, gene_output):
        """
        Forward pass of the module. Aggregates gene level features to gene set level features.

        :param gene_output: A tensor of shape (batch_size, num_genes, feature_dim), representing gene-level features.
        :return: A tensor of shape (batch_size, num_genesets, feature_dim), representing aggregated gene set level features.
        """
        batch_size = gene_output.size(0)
        num_genesets = len(self.genesets_indices)
        feature_dim = gene_output.size(2)
        aggregated_features = torch.zeros(
            batch_size, num_genesets, feature_dim, device=gene_output.device
        )

        for i, geneset in enumerate(self.genesets_indices):
            selected_features = gene_output[:, geneset, :]

            if self.pooling_type == "mean":
                aggregated_features[:, i, :] = torch.mean(selected_features, dim=1)
            elif self.pooling_type == "max":
                aggregated_features[:, i, :] = torch.max(selected_features, dim=1)[0]

        return aggregated_features



        
class GeneSetAttentionAggregator(nn.Module):
    def __init__(self, genesets_indices, softmax_mean=True):
        """
        Initializes the ParametricGeneSetAggregator module.
        :param gene_sets: A list of lists, where each inner list contains indices of genes in a gene set.

        # Example usage
        # Define gene sets
        # Example gene output tensor (batch_size, num_genes, feature_dim)
        >>> gene_output = torch.rand(256, 876, 32)  # Replace with actual gene output tensor
        >>> genesets_indices = [(0, 3, 10), (1, 3, 8), (10, 100, 101, 500)]  # Replace with actual gene sets
        >>> aggregator = ParametricGeneSetAggregator(genesets_indices)
        # Get gene set level features
        >>> gene_set_level_features = aggregator(gene_output)
        >>> gene_set_level_features.shape  # This should be (batch_size, num_gene_sets, feature_dim)
        """
        super(GeneSetAttentionAggregator, self).__init__()

        self.genesets_indices = genesets_indices
        self.softmax_mean = softmax_mean
        # Attention weights for each gene in each gene set
        self.attention_weights = nn.ParameterDict(
            {
                f"geneset_{i}": nn.Parameter(torch.randn(len(gene_set), 1))
                for i, gene_set in enumerate(genesets_indices)
            }
        )

    def forward(self, gene_features):
        """
        Forward pass of the module.

        :param gene_features: A tensor of shape (batch_size, num_genes, feature_dim), representing gene-level features.
        :return: A tensor of shape (batch_size, num_gene_sets, feature_dim), representing gene set level features.
        """
        batch_size = gene_features.size(0)
        gene_set_features = []

        for i, gene_set in enumerate(self.genesets_indices):
            set_features = gene_features[
                :, gene_set, :
            ]  # Extract features for genes in the set

            if self.softmax_mean:
                attention = F.softmax(
                    self.attention_weights[f"geneset_{i}"].expand(batch_size, -1, -1),
                    dim=1,
                )
                weighted_features = set_features * attention
                aggregated_features = torch.sum(weighted_features, dim=1)

            else:
                attention = self.attention_weights[f"geneset_{i}"]
                weighted_features = set_features * attention
                aggregated_features = torch.mean(weighted_features, dim=1)

            gene_set_features.append(aggregated_features)

        gene_set_features = torch.stack(gene_set_features, dim=1)

        return gene_set_features


class GeneSetAggregator(nn.Module):
    def __init__(self, genesets_indices, mode="attention", pooling_type="mean"):
        """
        Initializes the CellPathwayAggregator module.
        :param genesets_indices: A list of lists, each sublist contains indices of genes for a specific gene set.
        :param mode: The aggregation mode ('pooling' or 'attention').
        :param pooling_type: The type of pooling ('mean' or 'max'), used if mode is 'pooling'.

        Example
        ===============
        >>> gene_output = torch.rand(10, 876, 32)  # Replace with actual gene output tensor
        >>> genesets_indices = [(0, 3, 10), (1, 3, 8), (10, 100, 101, 500)]  # Replace with actual gene sets
        >>> aggregator = GeneSetAggregator(genesets_indices, mode='attention')
        >>> aggregator(gene_output).shape

        """
        super(GeneSetAggregator, self).__init__()
        self.mode = mode
        self.genesets_indices = genesets_indices

        if mode == "pooling":
            self.aggregator = GeneSetPoolingAggregator(genesets_indices, pooling_type)
        elif mode == "attention":
            self.aggregator = GeneSetAttentionAggregator(genesets_indices)
        else:
            raise ValueError("Invalid mode. Use 'pooling' or 'attention'.")

    def forward(self, gene_features):
        """
        Forward pass of the module.
        :param gene_features: A tensor of shape (batch_size, num_genes, feature_dim).
        :return: A tensor of shape (batch_size, num_cellpathways).
        """
        geneset_feats = self.aggregator(gene_features)
        return geneset_feats



import torch
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F
from torch.nn import Sequential, Linear, ModuleList, ReLU, Dropout
from copy import deepcopy
import pandas as pd
import numpy as np


import random


def fixseed(seed=42):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



class ClassDecoder(nn.Module):
    def __init__(
        self,
        input_dim=44,
        dense_layers=[],
        out_dim=2,
        dropout_p=0.0,
        batch_norms=True,
        seed=42,
    ):
        super().__init__()
        self.seed = seed
        fixseed(seed=seed)

        self.batch_norms = batch_norms  
        self.feature_gate = nn.Parameter(torch.zeros(input_dim))  # logits，
        self.linear = nn.Linear(input_dim, out_dim, bias=True)

        self.use_residual = True
        hidden = dense_layers[-1] if (dense_layers is not None and len(dense_layers) > 0) else input_dim
        self.residual = nn.Sequential(
            nn.Linear(input_dim, hidden, bias=True),
            nn.Tanh(),                 # use tanh instead of ReLU
            nn.Dropout(dropout_p),
            nn.Linear(hidden, out_dim, bias=True),
        )
        self.res_scale = nn.Parameter(torch.tensor(0.0)) 

        # temperature
        self.log_temperature = nn.Parameter(torch.log(torch.tensor(1.0)))

    def forward(self, x):
        gate = 2 * torch.sigmoid(self.feature_gate)  # (44,)
        xg = x * gate

        y = self.linear(xg)
        if self.use_residual:
            y = y + torch.tanh(self.res_scale) * self.residual(xg)

        temperature = torch.exp(self.log_temperature)
        return y / temperature


class ProtoNetDecoder(nn.Module):
    def __init__(
        self,
        input_dim,
        out_dim=2,
        dense_layers=[],
        dropout_p=0.0,      
        batch_norms=True, 
        seed=42,
    ):
        super(ProtoNetDecoder, self).__init__()

        self.seed = seed
        fixseed(seed=seed)

        # ===== layer dims (interface preserved) =====
        _dense_layers = [input_dim]
        _dense_layers.extend(dense_layers)
        self._dense_layers = _dense_layers

        # ===== linear projection (optional, very light) =====
        self.lins = ModuleList()
        for i in range(len(_dense_layers) - 1):
            self.lins.append(
                Linear(_dense_layers[i], _dense_layers[i + 1], bias=False)
            )

        # ===== prototype classifier =====
        last_hidden = _dense_layers[-1]
        self.W = nn.Parameter(torch.Tensor(out_dim, last_hidden))
        self.b = nn.Parameter(torch.zeros(out_dim))

        init.normal_(self.W, mean=0.0, std=0.01)

        # ===== temperature (kept but frozen) =====
        self.log_temperature = nn.Parameter(
            torch.zeros(1), requires_grad=False
        )  # temperature = 1

        self.num_classes = out_dim
        self.feature_dim = input_dim

    def forward(self, x):
        """
        x: [B, input_dim]
        return: logits [B, out_dim]  (NO softmax, PRC-safe)
        """

        # optional light projection (if dense_layers != [])
        for lin in self.lins:
            x = lin(x)  # no activation, no BN, no dropout

        # L2 normalize features & prototypes
        x = F.normalize(x, p=2, dim=1)
        W = F.normalize(self.W, p=2, dim=1)

        # cosine similarity
        logits = torch.matmul(x, W.t()) + self.b

        temperature = torch.exp(self.log_temperature)  # = 1
        return logits / temperature

    def initialize_parameters(self, support_features, support_labels):
        """
        Initialize prototypes with class means (ProtoNet-style).

        support_features: [N, D]
        support_labels:   [N, C] (one-hot)
        """
        with torch.no_grad():
            for i in range(self.num_classes):
                mask = support_labels[:, i] == 1
                if mask.sum() == 0:
                    continue
                class_mean = support_features[mask].mean(dim=0)
                self.W[i].copy_(class_mean)
                


# # Define the Prototypical Network without fine-tuning
class ProtoNetNFTDecoder:

    def __init__(self, temperature=1e-3):

        self.prototype_class_map = {
            "PD": 0,
            "SD": 0,
            "PR": 1,
            "CR": 1,
            1: 1,
            0: 0,
            "R": 1,
            "NR": 0,
        }
        self.temperature = temperature

        
    def fit(self, support_set):
        """
        support_set: the last column is the RECIST label column
        """
        self.label_col = support_set.columns[-1]
        self.feature_col = support_set.columns[:-1]

        unique_recist_labels = support_set[self.label_col].unique()
        out_recist = set(unique_recist_labels) - set(self.prototype_class_map.keys())
        assert len(out_recist) == 0, "Unepxected RECIST labels: %s" % out_recist

        prototype_features = support_set.groupby(self.label_col).median()
        prototype_types = prototype_features.index
        prototype_representation = torch.tensor(prototype_features.values)
        prototype_representation = F.normalize(prototype_representation, p=2, dim=1)
        self.prototype_representation = prototype_representation
        self.prototype_types = prototype_types
        return self

    def transform(self, query_set):
        label_col = query_set.columns[-1]
        assert label_col == self.label_col, "%s is missing!" % self.label_col
    
        X = torch.tensor(query_set[self.feature_col].values, dtype=torch.float)
        X = F.normalize(X, dim=1)
        proto_mean = F.normalize(
            self.prototype_representation.mean(dim=0), dim=0
        )
        proj = (X @ proto_mean.unsqueeze(1)) * proto_mean.unsqueeze(0)
        X = F.normalize(X - proj, dim=1)

        similarities = torch.mm(
            X, self.prototype_representation.T
        )
        similarities = similarities / self.temperature
    
        probabilities = F.softmax(similarities, dim=1)
        probabilities = probabilities.detach().numpy()
    
        dfprob = pd.DataFrame(
            probabilities, index=query_set.index, columns=self.prototype_types
        )
    
        dfprob2 = dfprob.copy()
        dfprob2.columns = dfprob2.columns.map(self.prototype_class_map)
        dfprob2 = dfprob2.T.reset_index().groupby(self.label_col).sum().T
        return dfprob2



class RegDecoder(nn.Module):
    def __init__(
        self,
        input_dim=32,
        dense_layers=[],
        out_dim=1,
        dropout_p=0.0,
        batch_norms=True,
        seed=42,
    ):
        """
        Regression
        """
        super(RegDecoder, self).__init__()
        self.seed = seed
        fixseed(seed=seed)

        ## Input
        self.input_norm = torch.nn.BatchNorm1d(input_dim)

        _dense_layers = [input_dim]
        _dense_layers.extend(dense_layers)
        self._dense_layers = _dense_layers
        self.batch_norms = batch_norms

        ## Dense
        self.lins = ModuleList()
        for i in range(len(_dense_layers) - 1):
            lin = Linear(_dense_layers[i], _dense_layers[i + 1])
            self.lins.append(lin)

        ## Batchnorm
        self._batch_norms = ModuleList()
        for i in range(len(_dense_layers) - 1):
            self._batch_norms.append(
                deepcopy(torch.nn.BatchNorm1d(_dense_layers[i + 1]))
            )

        ## Dropout
        self.dropout = nn.Dropout(dropout_p)

        # Output layer
        last_hidden = _dense_layers[-1]
        self.out = Linear(last_hidden, out_dim)

    def forward(self, x):
        if self.batch_norms & (len(self._batch_norms) == 0):
            x = self.input_norm(x)

        for lin, norm in zip(self.lins, self._batch_norms):
            if self.batch_norms:
                x = self.dropout(F.relu(norm(lin(x)), inplace=False))
            else:
                x = self.dropout(F.relu(lin(x), inplace=False))
        y = self.out(x)
        return y



import torch
import os
import tempfile
import gdown
from tqdm import tqdm


def download_with_progress(url, output_path):
    """
    Download a file from a URL with a progress bar using gdown.

    Args:
        url (str): The URL to download the file from.
        output_path (str): The local path to save the downloaded file.
    """
    print(f"Downloading model from {url}...")
    gdown.download(url, output_path, quiet=False)
    print(f"Model downloaded to: {output_path}")


def loadcompass(file, **kwargs):
    """
    Load a model from a file or URL.

    Args:
        file (str): Path to the model file or a URL.
        **kwargs: Additional arguments for torch.load.

    Returns:
        model: Loaded PyTorch model.
    """
    # Check if the file is a URL
    if file.startswith("http://") or file.startswith("https://"):
        # Create a temporary file to store the downloaded model
        temp_file_path = tempfile.mktemp()
        download_with_progress(file, temp_file_path)
        file = temp_file_path

    # Load the model
    model = torch.load(file, **kwargs)

    # Clean up wandb settings if present
    if hasattr(model, "with_wandb") and model.with_wandb:
        model.wandb._settings = ""

    # Update device if map_location is 'cpu'
    if kwargs.get("map_location") == "cpu":
        model.device = "cpu"

    # Remove temporary file if downloaded
    if file.startswith(tempfile.gettempdir()):
        os.remove(file)

    return model



# -*- coding: utf-8 -*-
"""
Created on Fri Nov  3 13:31:25 2023

@author: Wanxiang Shen

"""

import torch.nn as nn
from torch.nn.modules.container import ModuleList
from torchvision.ops import MLP

import copy, torch
from .layer import CosformerLayer, PerformerLayer, FlowformerLayer
from .layer import VanillaTransformerLayer, FlashTransformerLayer
from ..embedder import GeneEmbedding, _GeneInitialization

from .layer.norm import create_norm


def _get_clones(module, N):
    # FIXME: copy.deepcopy() is not defined on nn.module
    return ModuleList([copy.deepcopy(module) for i in range(N)])


class Encoder(nn.Module):

    def __init__(
        self,
        encoder_type="transformer",
        d_model=32,
        dim_feedforward=64,
        nhead=2,
        num_layers=1,
        dropout=0,
        **kwargs,
    ):

        super(Encoder, self).__init__()

        self.encoder_type = encoder_type
        self.d_model = d_model
        self.nhead = nhead
        self.num_layers = num_layers
        self.dim_head = d_model // nhead  # 16
        self.dropout = dropout

        if encoder_type == "cosformer":
            encoder_layer = CosformerLayer(
                embed_dim=d_model, num_heads=nhead, dropout=dropout, **kwargs
            )
        elif encoder_type == "performer":
            encoder_layer = PerformerLayer(
                embed_dim=d_model, num_heads=nhead, dropout=dropout, **kwargs
            )
        elif encoder_type == "Vanillatransformer":
            encoder_layer = VanillaTransformerLayer(
                embed_dim=d_model, num_heads=nhead, dropout=dropout, **kwargs
            )
        elif encoder_type == "flowformer":
            encoder_layer = FlowformerLayer(
                embed_dim=d_model, num_heads=nhead, dropout=dropout, **kwargs
            )
        elif encoder_type == "transformer":
            encoder_layer = nn.TransformerEncoderLayer(
                d_model=d_model,
                nhead=nhead,
                dropout=dropout,
                dim_feedforward=dim_feedforward,
                batch_first=True,
                **kwargs,
            )

        elif encoder_type == "flashformer":
            encoder_layer = FlashTransformerLayer(
                d_model=d_model,
                nhead=nhead,
                dropout=dropout,
                dim_feedforward=dim_feedforward,
                batch_first=True,
                **kwargs,
            )

        else:
            raise NotImplementedError(
                f"Not implemented transformer type: {encoder_type}"
            )

        self.layers = _get_clones(encoder_layer, num_layers)

    def forward(self, x, output_attentions=False):

        att_list = []
        for l in range(self.num_layers):
            if self.encoder_type == "transformer":
                x = self.layers[l](x)
                att = None
            else:
                x, att = self.layers[l](x, output_attentions=output_attentions)
            att_list.append(att)

        return x, att_list


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        num_cancer_types=33,
        encoder_type="transformer",
        input_dim=15672,
        nhead=2,
        d_model=32,
        num_layers=2,
        dropout=0.0,
        dim_feedforward=32,
        pos_emb="learnable",
        **kwargs,
    ):
        """
        encoder_type: {'transformer', 'reformer', 'performer'}
        pos_emb: {None, 'learnable', 'gene2vect', 'umap'}
        d_model: {16,32,64,128,512}
        """
        super().__init__()
        self.encoder_type = encoder_type
        self.num_cancer_types = num_cancer_types
        self.input_dim = input_dim
        self.nhead = nhead
        self.d_model = d_model
        self.num_layers = num_layers
        self.dropout = dropout
        self.dim_feedforward = dim_feedforward
        self.pos_emb = pos_emb

        self.pid_token_embedder = nn.Parameter(torch.randn(1, d_model))
        initialization_ = _GeneInitialization.from_str("uniform")
        initialization_.apply(self.pid_token_embedder, d_model)

        self.cancer_token_embedder = nn.Embedding(num_cancer_types, d_model)
        self.gene_token_embedder = GeneEmbedding(input_dim, d_model, pos_emb)

        self.encoder = Encoder(
            encoder_type=encoder_type,
            d_model=d_model,
            dropout=dropout,
            dim_feedforward=dim_feedforward,
            nhead=nhead,
            num_layers=num_layers,
            **kwargs,
        )

    def forward(self, x, output_attentions=False):

        cancer_types = x[:, 0].long()  # first column is the cancer type
        genes = x[:, 1:]  #

        # expand patient/sample/dataset token (equals to the CLS token)
        pid_embed = self.pid_token_embedder.expand(x.size(0), -1).unsqueeze(1)  # B,1, C
        # convert cancer types
        cancer_embed = self.cancer_token_embedder(cancer_types).unsqueeze(1)  # B,1, C
        # gene embeddings
        gene_embed = self.gene_token_embedder(genes)  # B, L, C

        # concat dataset_embed,cancer_embed,
        transformer_input = torch.cat(
            [pid_embed, cancer_embed, gene_embed], dim=1
        )  # B, L+2, C

        x, attn = self.encoder(transformer_input, output_attentions=output_attentions)

        if output_attentions:
            return x, attn
        else:
            return x


class MLPEncoder(nn.Module):
    def __init__(self, input_dim, dense_layers=[512, 256, 128], dropout=0.2):
        super().__init__()
        self.encoder = MLP(
            in_channels=input_dim,
            hidden_channels=dense_layers,
            norm_layer=torch.nn.BatchNorm1d,
            dropout=dropout,
        )

    def forward(self, x):
        x = self.encoder(x)
        return x



# -*- coding: utf-8 -*-
"""
Created on Fri Nov  3 13:31:25 2023

@author: Wanxiang Shen

"""
import torch
import torch.nn as nn
from ..encoder import TransformerEncoder, MLPEncoder
from ..decoder import ClassDecoder, RegDecoder, ProtoNetDecoder
from ..projector import DisentangledProjector, EntangledProjector


import numpy as np
import random


def fixseed(seed=42):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



class Compass(nn.Module):

    def __init__(
        self,
        input_dim,
        task_dim,
        task_type,
        num_cancer_types=33,
        embed_dim=32,
        #### projections
        proj_disentangled=True,
        proj_level="cellpathway",
        proj_pid=False,
        proj_cancer_type=True,
        ref_for_task=True,
        encoder="transformer",
        encoder_dropout=0.0,
        task_dense_layer=[24],
        task_batch_norms=True,
        transformer_dim=32,
        transformer_num_layers=1,
        transformer_nhead=2,
        transformer_pos_emb="learnable",
        seed=42,
        **encoder_kwargs
    ):
        """
        input_dim:  number of tokens
        task_dim: supervised learning task dim
        task_type: {'r', 'c'}
        num_cancer_types: int, number cancer types, default 33.
        embed_dim: latent vector dim
        encoder: {'transfomer', 'flowformer', ...}
        task_dense_layer: dense layer of task
        transformer_pos_emb: {None, 'umap', 'pumap'}
        """
        super().__init__()

        self.input_dim = input_dim
        self.task_dim = task_dim
        self.task_type = task_type

        self.proj_disentangled = proj_disentangled
        self.proj_level = proj_level
        self.proj_pid = proj_pid
        self.proj_cancer_type = proj_cancer_type
        self.ref_for_task = ref_for_task

        self.num_cancer_types = num_cancer_types
        self.encoder = encoder
        self.encoder_dropout = encoder_dropout
        self.transformer_dim = transformer_dim
        self.transformer_num_layers = transformer_num_layers
        self.transformer_pos_emb = transformer_pos_emb
        self.transformer_nhead = transformer_nhead
        self.task_batch_norms = task_batch_norms
        self.task_dense_layer = task_dense_layer
        self.encoder_kwargs = encoder_kwargs
        self.seed = seed

        fixseed(seed=self.seed)

        self.inputencoder = TransformerEncoder(
            num_cancer_types=num_cancer_types,
            encoder_type=encoder,
            input_dim=input_dim,
            d_model=transformer_dim,
            num_layers=transformer_num_layers,
            nhead=transformer_nhead,
            dropout=encoder_dropout,
            pos_emb=transformer_pos_emb,
            **encoder_kwargs
        )

        if proj_disentangled:
            self.latentprojector = DisentangledProjector(
                input_dim,
                transformer_dim,
                proj_pid=proj_pid,
                proj_cancer_type=proj_cancer_type,
            )

            self.geneset_feature_name = self.latentprojector.geneset_proj_cols
            self.celltype_feature_name = self.latentprojector.cellpathway_proj_cols
            self.ref_gene_ids = self.latentprojector.ref_gene_ids
            self.ref_geneset_ids = self.latentprojector.ref_geneset_ids
            self.ref_celltype_ids = self.latentprojector.ref_celltype_ids

            if proj_level == "cellpathway":
                a = self.celltype_feature_name
                k = self.ref_celltype_ids
            else:
                a = self.geneset_feature_name
                k = self.ref_geneset_ids

            if not self.ref_for_task:
                self.embed_dim = len(a) - len(k)
                self.embed_feature_names = a
                self.proj_feature_names = [a[i] for i in range(len(a)) if i not in k]
            else:
                self.embed_dim = len(a)
                self.embed_feature_names = a
                self.proj_feature_names = a

        else:
            self.latentprojector = EntangledProjector(transformer_dim)
            self.embed_feature_names = range(len(embed_dim))
            self.embed_dim = embed_dim
            self.proj_feature_names = self.embed_feature_names

        model_args = {
            "input_dim": self.input_dim,
            "task_dim": self.task_dim,
            "task_type": self.task_type,
            "proj_level": self.proj_level,
            "proj_pid": self.proj_pid,
            "proj_cancer_type": self.proj_cancer_type,
            "proj_disentangled": self.proj_disentangled,
            "embed_dim": self.embed_dim,
            "num_cancer_types": self.num_cancer_types,
            "encoder": self.encoder,
            "encoder_dropout": self.encoder_dropout,
            "transformer_dim": self.transformer_dim,
            "transformer_nhead": self.transformer_nhead,
            "transformer_num_layers": self.transformer_num_layers,
            "transformer_pos_emb": self.transformer_pos_emb,
            "task_batch_norms": self.task_batch_norms,
            "task_dense_layer": self.task_dense_layer,
            "seed": self.seed,
        }

        model_args.update(encoder_kwargs)
        self.model_args = model_args

        ## regression task
        if task_type == "r":
            self.taskdecoder = RegDecoder(
                input_dim=self.embed_dim,
                dense_layers=task_dense_layer,
                out_dim=task_dim,
                batch_norms=task_batch_norms,
                seed=self.seed,
            )

        ## classification task
        elif task_type == "c":
            self.taskdecoder = ClassDecoder(
                input_dim=self.embed_dim,
                dense_layers=task_dense_layer,
                out_dim=task_dim,
                batch_norms=task_batch_norms,
                seed=self.seed,
            )

        # for softmax classifier
        elif task_type == "f":
            self.taskdecoder = ProtoNetDecoder(
                input_dim=self.embed_dim,
                out_dim=task_dim,
                dense_layers=task_dense_layer,
                batch_norms=task_batch_norms,
                seed=self.seed,
            )

    def forward(self, x):

        # output： B,L+2, (dataset:1, cancer:1, gene),C
        encoding = self.inputencoder(x)
        geneset_level_proj, cellpathway_level_proj = self.latentprojector(encoding)

        # task_inputs: only input the context-oriented features (for downstream task)
        # Embedding: embeddings for contrastive learning
        if self.proj_level == "geneset":
            embedding = geneset_level_proj  # B,L
            emb_ref = embedding[:, self.ref_geneset_ids]

            mask = torch.ones(embedding.shape[1], dtype=torch.bool)
            mask[self.ref_geneset_ids] = False
            emb_used = embedding[:, mask]

        elif self.proj_level == "cellpathway":
            embedding = cellpathway_level_proj  # B,L
            emb_ref = embedding[:, self.ref_celltype_ids]

            mask = torch.ones(embedding.shape[1], dtype=torch.bool)
            mask[self.ref_celltype_ids] = False
            emb_used = embedding[:, mask]
            # print(emb_used.shape)

        if self.ref_for_task:
            y = self.taskdecoder(embedding)
        else:
            y = self.taskdecoder(emb_used)

        gene_encoding = encoding[:, 2:, :]
        gene_ref = gene_encoding[:, self.ref_gene_ids, :]

        return (embedding, (gene_ref, emb_ref)), y



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 15:05:13 2023

@author: Wanxiang Shen
"""


import torch
import os
from copy import deepcopy


class SaveBestModel:
    """
    Class to save the best model while training. If the current epoch's
    validation loss is less than the previous least less, then save the
    model state.
    """

    def __init__(
        self,
        save_dir="./model",
        save_name="best_model.pt",
        best_valid_loss=float("inf"),
    ):
        self.best_valid_loss = best_valid_loss
        self.save_dir = save_dir
        self.save_name = save_name

        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)
        self.inMemorySave = {}

    def __call__(self, current_valid_loss, epoch, model, optimizer, scaler):
        if current_valid_loss < self.best_valid_loss:
            self.best_valid_loss = current_valid_loss
            # print(f"\nBest validation loss: {self.best_valid_loss}")
            # print(f"\nSaving best model for epoch: {epoch+1}\n")
            smodel = deepcopy(model)  # should be deepcopy
            self.inMemorySave.update(
                {
                    "epoch": epoch + 1,
                    "model_args": smodel.model_args,
                    "model_state_dict": smodel.state_dict(),
                    "optimizer_state_dict": optimizer.state_dict(),
                    "scaler": scaler,
                }
            )

    def save(self):
        print("Saving final model...")
        print(f"\nBest validation loss: {self.best_valid_loss}")
        print(f"\nSaving best model on epoch: {self.inMemorySave['epoch']}\n")
        torch.save(self.inMemorySave, os.path.join(self.save_dir, self.save_name))



import torch
import torch.nn as nn
from .transformer import AbstractTrasnformerLayer
from .norm import create_norm

## Core code for Flow-Attention, Please refer to each folder for corresponding experiments


class Flow_Attention(nn.Module):
    # flow attention in normal version
    def __init__(
        self, d_model, n_heads, drop_out=0.01, d_input=None, d_output=None, eps=1e-6
    ):
        super(Flow_Attention, self).__init__()
        self.n_heads = n_heads
        if d_input is None:
            d_input = d_model
        if d_output is None:
            d_output = d_model
        self.query_projection = nn.Linear(d_input, d_model)
        self.key_projection = nn.Linear(d_input, d_model)
        self.value_projection = nn.Linear(d_input, d_model)
        self.out_projection = nn.Linear(d_model, d_output)
        self.dropout = nn.Dropout(drop_out)
        self.eps = eps

    def kernel_method(self, x):
        return torch.sigmoid(x)

    def dot_product(self, q, k, v):
        kv = torch.einsum("nhld,nhlm->nhdm", k, v)
        qkv = torch.einsum("nhld,nhdm->nhlm", q, kv)
        return qkv

    def forward(self, x):
        ## input: B (L or S) D; output: B L D
        ## Note: queries, keys, values are not projected yet
        ## 1. Linear projection
        # print(x.shape)

        queries = keys = values = x
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        queries = self.query_projection(queries).view(B, L, self.n_heads, -1)
        keys = self.key_projection(keys).view(B, S, self.n_heads, -1)
        values = self.value_projection(values).view(B, S, self.n_heads, -1)
        queries = queries.transpose(1, 2)
        keys = keys.transpose(1, 2)
        values = values.transpose(1, 2)
        # 2. Non-negative projection
        queries = self.kernel_method(queries)
        keys = self.kernel_method(keys)
        ## 3. Flow-Attention
        # (1) Calculate incoming and outgoing flow
        sink_incoming = 1.0 / (
            torch.einsum(
                "nhld,nhd->nhl", queries + self.eps, keys.sum(dim=2) + self.eps
            )
        )
        source_outgoing = 1.0 / (
            torch.einsum(
                "nhld,nhd->nhl", keys + self.eps, queries.sum(dim=2) + self.eps
            )
        )
        # (2) conservation refine for source and sink
        conserved_sink = torch.einsum(
            "nhld,nhd->nhl",
            queries + self.eps,
            (keys * source_outgoing[:, :, :, None]).sum(dim=2) + self.eps,
        )
        conserved_source = torch.einsum(
            "nhld,nhd->nhl",
            keys + self.eps,
            (queries * sink_incoming[:, :, :, None]).sum(dim=2) + self.eps,
        )
        conserved_source = torch.clamp(
            conserved_source, min=-1.0, max=1.0
        )  # for stability
        # (3) Competition & Allocation
        sink_allocation = torch.sigmoid(
            conserved_sink * (float(queries.shape[2]) / float(keys.shape[2]))
        )
        source_competition = torch.softmax(conserved_source, dim=-1) * float(
            keys.shape[2]
        )
        # (4) dot product
        x = (
            self.dot_product(
                queries * sink_incoming[:, :, :, None],  # for value normalization
                keys,
                values * source_competition[:, :, :, None],
            )  # competition
            * sink_allocation[:, :, :, None]
        ).transpose(
            1, 2
        )  # allocation
        ## (5) Final projection
        x = x.reshape(B, L, -1)
        x = self.out_projection(x)
        x = self.dropout(x)
        return x


class Flow_Attention_Causal(nn.Module):
    # flow attention in causal version
    def __init__(
        self, d_model, n_heads, drop_out=0.05, d_input=None, d_output=None, eps=1e-6
    ):
        super(Flow_Attention_Causal, self).__init__()
        self.n_heads = n_heads
        if d_input is None:
            d_input = d_model
        if d_output is None:
            d_output = d_model
        self.query_projection = nn.Linear(d_input, d_model)
        self.key_projection = nn.Linear(d_input, d_model)
        self.value_projection = nn.Linear(d_input, d_model)
        self.out_projection = nn.Linear(d_model, d_output)
        self.dropout = nn.Dropout(drop_out)
        self.eps = eps

    def kernel_method(self, x):
        return torch.sigmoid(x)

    def causal_dot_product(self, q, k, v):
        kv = torch.einsum("nhld,nhlm->nhldm", k, v)
        kv = torch.cumsum(kv, dim=2)
        qkv = torch.einsum("nhld,nhldm->nhlm", q, kv)
        return qkv

    def forward(self, x):
        ## input: B (L or S) D; output: B L D
        ## Note: queries, keys, values are not projected yet
        queries = keys = values = x
        ## 1. Linear projection
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        queries = self.query_projection(queries).view(B, L, self.n_heads, -1)
        keys = self.key_projection(keys).view(B, S, self.n_heads, -1)
        values = self.value_projection(values).view(B, S, self.n_heads, -1)
        queries = queries.transpose(1, 2)
        keys = keys.transpose(1, 2)
        values = values.transpose(1, 2)
        # 2. Non-negative projection
        queries = self.kernel_method(queries)
        keys = self.kernel_method(keys)
        ## 3. Causal Flow-Attention
        # (1) Calculate incoming and outgoing flow
        sink_incoming = 1.0 / (
            torch.einsum(
                "nhld,nhld->nhl", queries + self.eps, keys.cumsum(dim=2) + self.eps
            )
        )
        source_outgoing = 1.0 / (
            torch.einsum(
                "nhld,nhld->nhl", keys + self.eps, queries.cumsum(dim=2) + self.eps
            )
        )
        # approximate normal conservation col and row by multiplying corresponding element number
        normal = (((torch.arange(queries.shape[2])).float() + 1.0)).to(queries.device)[
            None, None, :
        ]
        sink_incoming = sink_incoming * normal
        source_outgoing = source_outgoing * normal
        # (2) conservation refine for source and sink
        conserved_sink = (
            torch.einsum(
                "nhld,nhld->nhl",
                queries + self.eps,
                (keys * source_outgoing[:, :, :, None]).cumsum(dim=2) + self.eps,
            )
            / normal
        )
        conserved_source = (
            torch.einsum(
                "nhld,nhld->nhl",
                keys + self.eps,
                (queries * sink_incoming[:, :, :, None]).cumsum(dim=2) + self.eps,
            )
            / normal
        )
        conserved_source = torch.clamp(
            conserved_source, min=-1.0, max=1.0
        )  # for stability
        # (3) Competition & Allocation
        sink_allocation = torch.sigmoid(conserved_sink)
        conserved_source = torch.exp(conserved_source)
        source_competition = (
            conserved_source / conserved_source.cumsum(dim=-1)
        ) * normal
        # (4) Causal dot product
        x = (
            self.causal_dot_product(
                queries
                * (
                    sink_incoming[:, :, :, None] / normal[:, :, :, None]
                ),  # for value normalization
                keys,
                values * source_competition[:, :, :, None],
            )  # competition
            * sink_allocation[:, :, :, None]
        ).transpose(
            1, 2
        )  # allocation
        ## (5) Final projection
        x = x.reshape(B, L, -1)
        x = self.out_projection(x)
        x = self.dropout(x)
        return x


class FlowformerLayer(nn.Module, AbstractTrasnformerLayer):
    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        norm="layernorm",
        norm_first=True,
        causal=False,
    ):
        super(FlowformerLayer, self).__init__()
        if not causal:
            self.self_attn = Flow_Attention(embed_dim, num_heads)
        else:
            self.self_attn = Flow_Attention_Causal(embed_dim, num_heads)
        self._ff_block = nn.Sequential(
            nn.Linear(embed_dim, embed_dim * 2),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(embed_dim * 2, embed_dim),
            nn.Dropout(dropout),
        )
        self.dropout1 = nn.Dropout(dropout)
        self.norm1 = create_norm(norm, embed_dim)
        self.norm2 = create_norm(norm, embed_dim)
        self.norm_first = norm_first
        self.support_output_attentions = False

    def _sa_block(self, x):
        x = self.self_attn(x)
        return self.dropout1(x)

    def forward(self, x, attn_mask=None, output_attentions=False):
        assert (
            output_attentions == False
        ), "output_attentions not implemented for Cosformer"
        if self.norm_first:
            x = x + self._sa_block(self.norm1(x))
            x = x + self._ff_block(self.norm2(x))
        else:
            x = self.norm1(x + self._sa_block(x))
            x = self.norm2(x + self._ff_block(x))
        return x, None



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 16:44:09 2023

@author: Wanxiang Shen
"""
import torch
import torch.nn.functional as F
import torch.utils.data as Torchdata
import pandas as pd
import numpy as np
from tqdm import tqdm

tqdm.pandas(ascii=True)

from ..dataloader import GeneData
from .loss import entropy_regularization, independence_loss
from .loss import reference_consistency_loss


def worker_init_fn(worker_id):
    seed = torch.initial_seed() % 2**32
    np.random.seed(seed)


def r2_loss(output, target):
    """
    Computes the 1 - R^2 loss between the output and target tensors.

    Args:
        output (torch.Tensor): The predicted values by the model.
        target (torch.Tensor): The actual values.

    Returns:
        torch.Tensor: The 1 - R^2 loss.
    """
    output_mean = torch.mean(output)
    target_mean = torch.mean(target)
    cov = torch.sum((output - output_mean) * (target - target_mean))
    output_std = torch.sqrt(torch.sum((output - output_mean) ** 2))
    target_std = torch.sqrt(torch.sum((target - target_mean) ** 2))

    r = cov / (output_std * target_std)
    return 1 - r**2


def Adp_Trainer(train_loader, model, optimizer, tsk_loss, device, ctp_idx):

    model.train()
    total_loss = []

    # torch.autograd.set_detect_anomaly(True)
    # for data in tqdm(train_loader, ascii=True):
    for data in train_loader:

        triplet, label = data

        anchor_y_true, positive_y_true, negative_y_true = label
        anchor, positive, negative = triplet

        anchor = anchor.to(device)
        anchor_y_true = anchor_y_true.to(device)

        optimizer.zero_grad()

        (anchor_emb, anchor_refg), _ = model(anchor)

        y_pred = anchor_emb[:, [ctp_idx]]
        y_true = anchor_y_true  # torch.cat([anchor_y_true, positive_y_true, negative_y_true])

        # print(y_pred.shape, y_true.shape)

        loss = F.l1_loss(y_pred, y_true)

        # print(loss)

        loss.backward()
        optimizer.step()

        total_loss.append(loss.item())

    train_total_loss = np.mean(total_loss)

    return train_total_loss


@torch.no_grad()
def Adp_Tester(test_loader, model, optimizer, tsk_loss, device, ctp_idx):

    model.eval()
    total_loss = []

    for data in test_loader:
        triplet, label = data
        anchor, positive, negative = triplet
        anchor_y_true, positive_y_true, negative_y_true = label

        anchor = anchor.to(device)
        anchor_y_true = anchor_y_true.to(device)
        (anchor_emb, anchor_refg), _ = model(anchor)

        y_pred = anchor_emb[:, [ctp_idx]]
        y_true = anchor_y_true
        loss = F.l1_loss(y_pred, y_true)

        # print(y_pred, y_true)

        total_loss.append(loss.item())

    test_total_loss = np.mean(total_loss)

    return test_total_loss




# import gc
# import math
from typing import Dict, Mapping, Optional, Tuple, Any, Union

import torch
import numpy as np
from torch import nn, Tensor

# import torch.distributed as dist
import torch.nn.functional as F
from torch.nn import TransformerEncoder, TransformerEncoderLayer

from .norm import create_norm, create_activation
from .FlashMultiHead import FlashMHA


class FlashTransformerLayer(nn.Module):
    r"""FlashTransformerLayer is made up of self-attn and feedforward network.
    The class is modified from torch.nn.TransformerEncoderLayer to support the
    FlashAttention.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of intermediate layer, relu or gelu (default=relu).
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False``.

    Examples::
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8)
        >>> src = torch.rand(10, 32, 512)
        >>> out = encoder_layer(src)

    Alternatively, when ``batch_first`` is ``True``:
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8, batch_first=True)
        >>> src = torch.rand(32, 10, 512)
        >>> out = encoder_layer(src)
    """

    __constants__ = ["batch_first"]

    def __init__(
        self,
        d_model,
        nhead,
        dim_feedforward=128,
        dropout=0.1,
        activation="relu",
        layer_norm_eps=1e-5,
        norm="layernorm",  # batchnorm, layernorm, rmsnorm
        norm_first=True,
        batch_first=True,
        flash=False,
        device=None,
        dtype=None,
    ) -> None:

        super().__init__()

        self.flash = flash
        self.device = device
        self.dtype = dtype
        self.norm = norm

        self.self_attn = FlashMHA(
            embed_dim=d_model,
            num_heads=nhead,
            batch_first=batch_first,
            dropout=dropout,
            causal=False,
            device=device,
            dtype=dtype,
            flash=flash,
        )

        factory_kwargs = {"device": device, "dtype": dtype}
        # print(factory_kwargs)

        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward, **factory_kwargs)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model, **factory_kwargs)
        self.norm_first = norm_first

        self.norm1 = create_norm(norm, d_model)
        self.norm2 = create_norm(norm, d_model)

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = create_activation(activation)

    # 自注意力 block
    def _sa_block(self, x, need_weights=False):
        x, attn = self.self_attn(x, x, x, need_weights=need_weights)
        return self.dropout1(x), attn

    # 前馈网络 block
    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)

    def forward(self, x, output_attentions=False):
        if self.norm_first:
            x_prime, attn = self._sa_block(
                self.norm1(x), need_weights=output_attentions
            )
            x = x + x_prime
            x = x + self._ff_block(self.norm2(x))
        else:
            x_prime, attn = self._sa_block(x, need_weights=output_attentions)
            x = self.norm1(x + x_prime)
            x = self.norm2(x + self._ff_block(x))

        return x, attn



import torch
import torch.nn as nn
import torch.nn.functional as F


class PoolingAggregator(nn.Module):
    def __init__(self, agg_indices_series, pooling_type="mean"):
        """
        Initializes the PoolingAggregator module.
        :param agg_indices_series: A pd.Series, each_index is the name the aggregated features, each value is the list of the node to be aggreagated
        :param pooling_type: A string indicating the type of pooling ('mean' or 'max').
        """
        super(PoolingAggregator, self).__init__()
        self.agg_indices_series = agg_indices_series
        self.agg_indices = agg_indices_series.tolist()
        self.pooling_type = pooling_type

    def forward(self, gene_set_features):
        """
        Forward pass of the module.
        :param gene_set_features: A tensor of shape (batch_size, num_gene_sets).
        :return: A tensor of shape (batch_size, num_agg_features).

        """
        device = gene_set_features.device
        batch_size = gene_set_features.size(0)
        num_agg_features = len(self.agg_indices)
        aggregated_features = torch.zeros(batch_size, num_agg_features, device=device)

        for i, indices in enumerate(self.agg_indices):
            if self.pooling_type == "mean":
                aggregated_features[:, i] = torch.mean(
                    gene_set_features[:, indices], dim=1
                )
            elif self.pooling_type == "max":
                aggregated_features[:, i] = torch.max(
                    gene_set_features[:, indices], dim=1
                )[0]

        return aggregated_features


class AttentionAggregator(nn.Module):
    def __init__(self, agg_indices_series):
        """
        Initializes the AttentionAggregator module.

        :param agg_indices_series: A pd.Series, each_index is the name the aggregated features, each value is the list of the node to be aggreagated
        """
        super(AttentionAggregator, self).__init__()

        self.agg_indices_series = agg_indices_series
        self.agg_indices = agg_indices_series.tolist()

        # Attention weights for each cell type/pathway in each cellpathway set
        self.attention_weights = nn.ParameterDict(
            {
                f"{name}": nn.Parameter(torch.randn(len(getset_idx), 1))
                for name, getset_idx in agg_indices_series.items()
            }
        )

    def forward(self, gene_set_features):
        """
        Forward pass of the module. Aggregates cell type/pathway level features to cellpathway set level features using attention mechanism.
        :param gene_set_features: A tensor of shape (batch_size, num_gene_sets).
        :return: A tensor of shape (batch_size, num_cellpathway_sets), representing aggregated cellpathway set level features.
        """
        batch_size = gene_set_features.size(0)
        num_agg_features = len(self.agg_indices)
        aggregated_features = torch.zeros(
            batch_size, num_agg_features, device=gene_set_features.device
        )
        i = 0
        for name, getset_idx in self.agg_indices_series.items():
            set_features = gene_set_features[:, getset_idx]
            attention_scores = F.softmax(self.attention_weights[f"{name}"], dim=0)
            weighted_features = set_features * attention_scores.T
            aggregated_features[:, i] = torch.sum(weighted_features, dim=1)
            i += 1

        return aggregated_features
